#!/usr/bin/env python
"""
_WorkQueue_t_

WorkQueue tests
"""
from __future__ import print_function

from builtins import next, range
from future.utils import viewitems

import os
import threading
import time
import unittest
import logging

from retry import retry

from Utils.PythonVersion import PY3

from WMCore.WMBase import getTestBase
from WMCore.ACDC.DataCollectionService import DataCollectionService
from WMCore.Configuration import Configuration
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.File import File as WMFile
from WMCore.DataStructs.Run import Run
from WMCore.Lexicon import sanitizeURL
from WMCore.ResourceControl.ResourceControl import ResourceControl
from WMCore.Services.DBS.DBSErrors import DBSReaderError
from WMCore.Services.UUIDLib import makeUUID
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue as WorkQueueService
from WMCore.WMBS.Job import Job
from WMCore.WMSpec.StdSpecs.DQMHarvest import DQMHarvestWorkloadFactory
from WMCore.WMSpec.StdSpecs.StepChain import StepChainWorkloadFactory
from WMCore.WMSpec.StdSpecs.ReReco import ReRecoWorkloadFactory
from WMCore.WMSpec.WMWorkload import WMWorkload, WMWorkloadHelper
from WMCore.WorkQueue.WorkQueue import WorkQueue, globalQueue, localQueue
from WMCore.WorkQueue.WorkQueueExceptions import (WorkQueueWMSpecError, WorkQueueNoMatchingElements,
                                                  WorkQueueNoWorkError)
from WMCore.WorkQueue.DataStructs.WorkQueueElement import STATES
from WMQuality.Emulators import EmulatorSetup
from WMQuality.Emulators.DataBlockGenerator import Globals
from WMQuality.Emulators.RucioClient.MockRucioApi import PILEUP_DATASET
from WMQuality.Emulators.WMSpecGenerator.WMSpecGenerator import createConfig

from WMCore_t.WMSpec_t.samples.MultiTaskProductionWorkload \
    import workload as MultiTaskProductionWorkload
from WMCore_t.WorkQueue_t.WorkQueueTestCase import WorkQueueTestCase


NBLOCKS_HICOMM = 47
NFILES_HICOMM = 72
NBLOCKS_COSMIC = 58
NFILES_COSMIC = 108
NFILES_COSMICRAW = 141
TOTAL_EVENTS=10000


def rerecoWorkload(workloadName, arguments, assignArgs=None):
    factory = ReRecoWorkloadFactory()
    wmspec = factory.factoryWorkloadConstruction(workloadName, arguments)
    if assignArgs:
        args = factory.getAssignTestArguments()
        args.update(assignArgs)
        wmspec.updateArguments(args)
    return wmspec


def stepchainWorkload(workloadName, arguments):
    factory = StepChainWorkloadFactory()
    wmspec = factory.factoryWorkloadConstruction(workloadName, arguments)
    return wmspec


def getFirstTask(wmspec):
    """Return the 1st top level task"""
    return next(wmspec.taskIterator())


def syncQueues(queue, skipWMBS=False):
    """Sync parent & local queues and split work
        Workaround having to wait for couchdb replication and splitting polling
    """
    queue.backend.forceQueueSync()
    time.sleep(1)
    work = queue.processInboundWork()
    queue.performQueueCleanupActions(skipWMBS=skipWMBS)
    # after replication need to wait a while to update result
    time.sleep(3)
    return work


class WorkQueueTest(WorkQueueTestCase):
    """
    _WorkQueueTest_

    For /MinimumBias/ComissioningHI-v1/RAW the dataset has 47 blocks with 72 files.
    The Rucio emulator sets the block locations like:
        17 at 'T2_XX_SiteA', 'T2_XX_SiteB', 'T2_XX_SiteC'
        19 at 'T2_XX_SiteA', 'T2_XX_SiteB'
        11 at 'T2_XX_SiteA' only
    """

    def __init__(self, methodName='runTest'):
        super(WorkQueueTest, self).__init__(methodName=methodName, mockDBS=True, mockRucio=True)
        self.queueParams = {}
        self.queueParams['log_reporter'] = "WorkQueue_Unittest"
        self.queueParams['rucioAccount'] = "wma_test"
        self.queueParams['rucioAuthUrl'] = "http://cms-rucio-int.cern.ch"
        self.queueParams['rucioUrl'] = "https://cms-rucio-auth-int.cern.ch"


    def setupConfigCacheAndAgrs(self):
        self.rerecoArgs = ReRecoWorkloadFactory.getTestArguments()
        self.rerecoArgs["CouchDBName"] = self.configCacheDB
        self.rerecoArgs["ConfigCacheID"] = createConfig(self.rerecoArgs["CouchDBName"])

        self.mcArgs = StepChainWorkloadFactory.getTestArguments()
        self.mcArgs["CouchDBName"] = self.configCacheDB
        self.mcArgs['Step1']["ConfigCacheID"] = createConfig(self.mcArgs["CouchDBName"])

        self.parentProcArgs = ReRecoWorkloadFactory.getTestArguments()
        self.parentProcArgs.update(IncludeParents="True")
        self.parentProcArgs.update(InputDataset="/Cosmics/ComissioningHI-PromptReco-v1/RECO")
        self.parentProcArgs["CouchDBName"] = self.configCacheDB
        self.parentProcArgs["ConfigCacheID"] = createConfig(self.parentProcArgs["CouchDBName"])

        self.openRunningProcArgs = ReRecoWorkloadFactory.getTestArguments()
        self.openRunningProcArgs.update(OpenRunningTimeout=10)
        self.openRunningProcArgs["CouchDBName"] = self.configCacheDB
        self.openRunningProcArgs["ConfigCacheID"] = createConfig(self.openRunningProcArgs["CouchDBName"])

        self.pileupArgs = StepChainWorkloadFactory.getTestArguments()
        self.pileupArgs['Step1'].update(MCPileup=PILEUP_DATASET)
        self.pileupArgs['Step1'].update(InputDataset="/MinimumBias/ComissioningHI-v1/RAW",
                                        RequestNumEvents=TOTAL_EVENTS,
                                        SplittingAlgo="EventAwareLumiBased")
        self.pileupArgs["CouchDBName"] = self.configCacheDB
        self.pileupArgs['Step1']["ConfigCacheID"] = createConfig(self.pileupArgs["CouchDBName"])

        self.pileupMcArgs = StepChainWorkloadFactory.getTestArguments()
        self.pileupMcArgs['Step1'].update(MCPileup=PILEUP_DATASET)
        self.pileupArgs['Step1'].update(RequestNumEvents=TOTAL_EVENTS)
        self.pileupMcArgs["CouchDBName"] = self.configCacheDB
        self.pileupMcArgs['Step1']["ConfigCacheID"] = createConfig(self.pileupMcArgs["CouchDBName"])

    def setUp(self):
        """
        If we dont have a wmspec file create one
        """
        # undo any customizations
        Globals.GlobalParams.resetParams()

        # set up WMAgent config file for couchdb
        self.configFile = EmulatorSetup.setupWMAgentConfig()

        WorkQueueTestCase.setUp(self)
        self.setupConfigCacheAndAgrs()

        # Basic production Spec
        self.spec = stepchainWorkload('testProduction', self.mcArgs)
        self.spec.setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        getFirstTask(self.spec).addProduction(totalEvents=TOTAL_EVENTS)
        self.spec.setSpecUrl(os.path.join(self.workDir, 'testworkflow.spec'))
        self.spec.save(self.spec.specUrl())

        # Production spec plus pileup
        self.productionPileupSpec = stepchainWorkload('testProduction', self.pileupMcArgs)
        self.productionPileupSpec.setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        getFirstTask(self.productionPileupSpec).addProduction(totalEvents=TOTAL_EVENTS)
        self.productionPileupSpec.setSpecUrl(os.path.join(self.workDir, 'testworkflowPileupMc.spec'))
        self.productionPileupSpec.save(self.productionPileupSpec.specUrl())

        # Processing spec plus pileup
        self.processingPileupSpec = stepchainWorkload('testProcessing', self.pileupArgs)
        self.processingPileupSpec.setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB', 'T2_XX_SiteC'])
        getFirstTask(self.processingPileupSpec).addProduction(totalEvents=TOTAL_EVENTS)
        self.processingPileupSpec.setSpecUrl(os.path.join(self.workDir, 'testworkflowPileup.spec'))
        self.processingPileupSpec.save(self.processingPileupSpec.specUrl())

        # ReReco spec with whitelist
        self.whitelistSpec = rerecoWorkload('whitelistlistSpec', self.rerecoArgs)
        self.whitelistSpec.setSpecUrl(os.path.join(self.workDir,
                                                   'testWhitelist.spec'))
        getFirstTask(self.whitelistSpec).data.constraints.sites.whitelist = ['T2_XX_SiteB']
        self.whitelistSpec.save(self.whitelistSpec.specUrl())

        # ReReco spec with delay for running open
        self.openRunningSpec = rerecoWorkload('openRunningSpec', self.openRunningProcArgs)
        self.openRunningSpec.setSpecUrl(os.path.join(self.workDir,
                                                     'testOpenRunningSpec.spec'))
        self.openRunningSpec.save(self.openRunningSpec.specUrl())

        # Create queues
        globalCouchUrl = "%s/%s" % (self.testInit.couchUrl, self.globalQDB)
        logdbCouchUrl = "%s/%s" % (self.testInit.couchUrl, self.logDBName)
        reqdbUrl = "%s/%s" % (self.testInit.couchUrl, self.requestDBName)
        self.globalQueue = globalQueue(DbName=self.globalQDB,
                                       InboxDbName=self.globalQInboxDB,
                                       QueueURL=globalCouchUrl,
                                       central_logdb_url=logdbCouchUrl,
                                       UnittestFlag=True,
                                       RequestDBURL=reqdbUrl,
                                       **self.queueParams)
        #        self.midQueue = WorkQueue(SplitByBlock = False, # mid-level queue
        #                            PopulateFilesets = False,
        #                            ParentQueue = self.globalQueue,
        #                            CacheDir = None)
        # ignore mid queue as it causes database duplication's
        # copy jobStateMachine couchDB configuration here since we don't want/need to pass whole configuration
        jobCouchConfig = Configuration()
        jobCouchConfig.section_("JobStateMachine")
        jobCouchConfig.JobStateMachine.couchurl = os.environ["COUCHURL"]
        jobCouchConfig.JobStateMachine.couchDBName = "testcouchdb"
        jobCouchConfig.JobStateMachine.jobSummaryDBName = "wmagent_summary_test"
        jobCouchConfig.JobStateMachine.summaryStatsDBName = "stat_summary_test"
        # copy bossAir configuration here since we don't want/need to pass whole configuration
        bossAirConfig = Configuration()
        bossAirConfig.section_("BossAir")
        bossAirConfig.BossAir.pluginDir = "WMCore.BossAir.Plugins"
        bossAirConfig.BossAir.pluginNames = ["MockPlugin"]
        bossAirConfig.BossAir.section_("MockPlugin")
        bossAirConfig.BossAir.MockPlugin.fakeReport = os.path.join(getTestBase(),
                                                            'WMComponent_t/JobAccountant_t/fwjrs',
                                                            "MergeSuccess.pkl")
        bossAirConfig.section_("Agent")
        bossAirConfig.Agent.agentName = "TestAgent"
        bossAirConfig.section_("JobStateMachine")
        bossAirConfig.JobStateMachine.couchurl = os.environ["COUCHURL"]
        bossAirConfig.JobStateMachine.couchDBName = "testcouchdb"
        bossAirConfig.JobStateMachine.jobSummaryDBName = "wmagent_summary_test"
        bossAirConfig.JobStateMachine.summaryStatsDBName = "stat_summary_test"

        self.localQueue = localQueue(DbName=self.localQDB,
                                     InboxDbName=self.localQInboxDB,
                                     ParentQueueCouchUrl=globalCouchUrl,
                                     ParentQueueInboxCouchDBName=self.globalQInboxDB,
                                     JobDumpConfig=jobCouchConfig,
                                     BossAirConfig=bossAirConfig,
                                     CacheDir=self.workDir,
                                     central_logdb_url=logdbCouchUrl,
                                     RequestDBURL=reqdbUrl,
                                     **self.queueParams)

        self.localQueue2 = localQueue(DbName=self.localQDB2,
                                      InboxDbName=self.localQInboxDB2,
                                      ParentQueueCouchUrl=globalCouchUrl,
                                      ParentQueueInboxCouchDBName=self.globalQInboxDB,
                                      JobDumpConfig=jobCouchConfig,
                                      BossAirConfig=bossAirConfig,
                                      CacheDir=self.workDir,
                                      central_logdb_url=logdbCouchUrl,
                                      RequestDBURL=reqdbUrl,
                                      **self.queueParams)

        # configuration for the Alerts messaging framework, work (alerts) and
        # control  channel addresses to which alerts
        # these are destination addresses where AlertProcessor:Receiver listens
        config = Configuration()
        config.section_("Alert")
        config.Alert.address = "tcp://127.0.0.1:5557"
        config.Alert.controlAddr = "tcp://127.0.0.1:5559"

        # standalone queue for unit tests
        self.queue = WorkQueue(JobDumpConfig=jobCouchConfig,
                               BossAirConfig=bossAirConfig,
                               DbName=self.queueDB,
                               InboxDbName=self.queueInboxDB,
                               CacheDir=self.workDir,
                               config=config,
                               central_logdb_url=logdbCouchUrl,
                               RequestDBURL=reqdbUrl,
                               **self.queueParams)

        # create relevant sites in wmbs
        rc = ResourceControl()
        site_se_mapping = {'T2_XX_SiteA': 'T2_XX_SiteA', 'T2_XX_SiteB': 'T2_XX_SiteB'}
        for site, se in viewitems(site_se_mapping):
            rc.insertSite(site, 100, 200, se, cmsName=site, plugin="MockPlugin")
            daofactory = DAOFactory(package="WMCore.WMBS",
                                    logger=threading.currentThread().logger,
                                    dbinterface=threading.currentThread().dbi)
            addLocation = daofactory(classname="Locations.New")
            addLocation.execute(siteName=site, pnn=se)

        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def setupReReco(self, assignArgs=None, **kwargs):
        # Sample Tier1 ReReco spec
        self.rerecoArgs.update(kwargs)
        processingSpec = rerecoWorkload('testProcessing', self.rerecoArgs, assignArgs=assignArgs)
        processingSpec.setSpecUrl(os.path.join(self.workDir, 'testProcessing.spec'))
        processingSpec.save(processingSpec.specUrl())
        return processingSpec

    def setupParentProcSpec(self, assignArgs=None, **kwargs):
        # Sample Tier1 ReReco spec with parent
        self.parentProcArgs.update(kwargs)
        parentProcSpec = rerecoWorkload('testParentProcessing', self.parentProcArgs, assignArgs=assignArgs)
        parentProcSpec.setSpecUrl(os.path.join(self.workDir, 'testParentProcessing.spec'))
        parentProcSpec.save(parentProcSpec.specUrl())
        return parentProcSpec

    def setupHighPrioReReco(self, assignArgs=None, **kwargs):
        # High priority ReReco spec
        self.rerecoArgs.update(kwargs)
        highPrioReReco = rerecoWorkload('highPrioSpec', self.rerecoArgs, assignArgs=assignArgs)
        highPrioReReco.data.request.priority = 999998
        highPrioReReco.setSpecUrl(os.path.join(self.workDir, 'highPrioSpec.spec'))
        highPrioReReco.save(highPrioReReco.specUrl())
        return highPrioReReco

    def tearDown(self):
        """tearDown"""
        super(WorkQueueTest, self).tearDown()
        # Delete WMBSAgent config file
        EmulatorSetup.deleteConfig(self.configFile)

    def createResubmitSpec(self, serverUrl, couchDB, parentage=False):
        """
        _createResubmitSpec_
        Create a bogus resubmit workload.
        """
        site = ["T1_US_FNAL"]
        workload = WMWorkloadHelper(WMWorkload("TestWorkload"))
        reco = workload.newTask("reco")
        workload.setOwnerDetails(name="evansde77", group="DMWM")
        workload.setSiteWhitelist(site)
        # first task uses the input dataset
        reco.addInputDataset(name="/PRIMARY/processed-v1/TIERONE",
                             primary="PRIMARY", processed="processed-v1", tier="TIERONE")
        reco.data.input.splitting.algorithm = "File"
        reco.data.input.splitting.include_parents = parentage
        reco.setTaskType("Processing")
        cmsRunReco = reco.makeStep("cmsRun1")
        cmsRunReco.setStepType("CMSSW")
        reco.applyTemplates()
        cmsRunRecoHelper = cmsRunReco.getTypeHelper()
        cmsRunRecoHelper.addOutputModule("outputRECO",
                                         primaryDataset="PRIMARY",
                                         processedDataset="processed-v2",
                                         dataTier="TIERTWO",
                                         lfnBase="/store/dunkindonuts",
                                         mergedLFNBase="/store/kfc")

        workload.setTrustLocationFlag(inputFlag=True, pileupFlag=False)
        dcs = DataCollectionService(url=serverUrl, database=couchDB)

        def getJob(workload):
            job = Job()
            job["task"] = workload.getTask("reco").getPathName()
            job["workflow"] = workload.name()
            job["location"] = site
            job["owner"] = workload.getOwner().get("name")
            job["group"] = workload.getOwner().get("group")
            return job

        testFileA = WMFile(lfn=makeUUID(), size=1024, events=1024, parents=['parent1'])
        testFileA.setLocation(site)
        testFileA.addRun(Run(1, 1, 2))
        testFileB = WMFile(lfn=makeUUID(), size=1024, events=1024, parents=['parent2'])
        testFileB.setLocation(site)
        testFileB.addRun(Run(1, 3, 4))
        testJobA = getJob(workload)
        testJobA.addFile(testFileA)
        testJobA.addFile(testFileB)

        dcs.failedJobs([testJobA])
        topLevelTask = workload.getTopLevelTask()[0]
        workload.truncate("Resubmit_TestWorkload", topLevelTask.getPathName(),
                          serverUrl, couchDB)

        return workload

    def testProduction(self):
        """
        Enqueue and get work for a production WMSpec.
        """
        specfile = self.spec.specUrl()
        numUnit = 1
        jobSlot = [10] * numUnit  # array of jobs per block
        total = sum(jobSlot)
        for _ in range(numUnit):
            self.queue.queueWork(specfile)
        self.assertEqual(numUnit, len(self.queue))

        # try to get work
        work = self.queue.getWork({'SiteDoesNotExist': jobSlot[0]}, {})
        self.assertEqual([], work)  # not in whitelist

        work = self.queue.getWork({'T2_XX_SiteA': 0}, {})
        self.assertEqual([], work)
        work = self.queue.getWork({'T2_XX_SiteA': jobSlot[0]}, {})
        self.assertEqual(len(work), 1)

        # no more work available
        self.assertEqual(0, len(self.queue.getWork({'T2_XX_SiteA': total}, {})))

    def testProductionMultiQueue(self):
        """Test production with multiple queueus"""
        specfile = self.spec.specUrl()
        numUnit = 1
        jobSlot = [10] * numUnit  # array of jobs per block
        total = sum(jobSlot)

        self.globalQueue.queueWork(specfile)
        self.assertEqual(numUnit, len(self.globalQueue))

        # pull work to localQueue2 - check local doesn't get any
        numWork = self.localQueue2.pullWork({'T2_XX_SiteA': total})
        self.assertEqual(numUnit, numWork)
        self.assertEqual(0, self.localQueue.pullWork({'T2_XX_SiteA': total}))
        syncQueues(self.localQueue)
        syncQueues(self.localQueue2)
        self.assertEqual(0, len(self.localQueue.status(status='Available')))
        self.assertEqual(numUnit, len(self.localQueue2.status(status='Available')))
        self.assertEqual(numUnit, len(self.globalQueue.status(status='Acquired')))
        self.assertEqual(sanitizeURL(self.localQueue2.params['QueueURL'])['url'],
                         self.globalQueue.status()[0]['ChildQueueUrl'])

    def testPriority(self):
        """
        Test priority change functionality
        """
        jobSlot = 10
        totalSlices = 1

        self.queue.queueWork(self.spec.specUrl())
        self.queue.processInboundWork()

        # priority change
        self.queue.setPriority(50, self.spec.name())
        # test elements are now cancelled
        self.assertEqual([x['Priority'] for x in self.queue.status(RequestName=self.spec.name())],
                         [50] * totalSlices)
        self.assertRaises(RuntimeError, self.queue.setPriority, 50, 'blahhhhh')

        # claim all work
        work = self.queue.getWork({'T2_XX_SiteA': jobSlot}, {})
        self.assertEqual(len(work), totalSlices)

        # no more work available
        self.assertEqual(0, len(self.queue.getWork({'T2_XX_SiteA': jobSlot}, {})))

    def testProcessing(self):
        """
        Enqueue and get work for a processing WMSpec.
        """
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB", "T2_XX_SiteC"]})
        specfile = processingSpec.specUrl()

        # Queue Work & check accepted
        self.queue.queueWork(specfile)
        self.queue.processInboundWork()
        self.assertEqual(NBLOCKS_HICOMM, len(self.queue))

        self.queue.updateLocationInfo()
        # No resources
        work = self.queue.getWork({}, {})
        self.assertEqual(len(work), 0)
        work = self.queue.getWork({'T2_XX_SiteA': 0, 'T2_XX_SiteB': 0}, {})
        self.assertEqual(len(work), 0)

        # Get the first bit of work available at site C
        work = self.queue.getWork({'T2_XX_SiteC': 1}, {})
        self.assertEqual(len(work), 1)  # Double check A
        processedBlocks = len(work)
        processedFiles = work[0]["NumOfFilesAdded"]

        # Get the rest the of work available at site C
        work = self.queue.getWork({'T2_XX_SiteC': 1000}, {})
        processedBlocks += len(work)
        for element in work:
            processedFiles += element["NumOfFilesAdded"]
        self.assertEqual(processedBlocks, 9)
        self.assertEqual(processedFiles, 14)

        # Get the rest the of work available at site B
        work = self.queue.getWork({'T2_XX_SiteB': 1000}, {})
        processedBlocks += len(work)
        for element in work:
            processedFiles += element["NumOfFilesAdded"]
        self.assertEqual(processedBlocks, 31)
        self.assertEqual(processedFiles, 52)

        # Make sure no work left for B or C
        work = self.queue.getWork({'T2_XX_SiteB': 1000, 'T2_XX_SiteC': 1000}, {})
        self.assertEqual(len(work), 0)

        # Make sure we get all the work when we include A
        work = self.queue.getWork({'T2_XX_SiteA': 1000}, {})
        processedBlocks += len(work)
        for element in work:
            processedFiles += element["NumOfFilesAdded"]
        self.assertEqual(processedBlocks, NBLOCKS_HICOMM)
        self.assertEqual(processedFiles, NFILES_HICOMM)

        # Make sure no remaining work
        work = self.queue.getWork({'T2_XX_SiteA': 1000, 'T2_XX_SiteB': 1000}, {})
        self.assertEqual(len(work), 0)  # no more work available

    def testBlackList(self):
        """
        Black list functionality
        """
        blacklistSpec = self.setupReReco(assignArgs={'SiteWhitelist':["T2_XX_SiteB"], 'SiteBlacklist': ["T2_XX_SiteA"]})
        specfile = blacklistSpec.specUrl()

        # Queue Work & check accepted
        self.queue.queueWork(specfile)
        self.queue.processInboundWork()
        self.assertEqual(NBLOCKS_HICOMM, len(self.queue))
        self.queue.updateLocationInfo()

        # T2_XX_SiteA is in blacklist, no work pulled
        work = self.queue.getWork({'T2_XX_SiteA': 1000}, {})
        self.assertEqual(len(work), 0)

        # T2_XX_SiteB can run most blocks
        work = self.queue.getWork({'T2_XX_SiteB': 1000}, {})
        self.assertEqual(len(work), 31)

    def testWhiteList(self):
        """
        White list functionality
        """
        specfile = self.whitelistSpec.specUrl()

        # Queue Work & check accepted
        self.queue.queueWork(specfile)
        self.queue.processInboundWork()
        self.assertEqual(NBLOCKS_HICOMM, len(self.queue))
        self.queue.updateLocationInfo()

        # Only SiteB in whitelist
        work = self.queue.getWork({'T2_XX_SiteA': 1000}, {})
        self.assertEqual(len(work), 0)

        # Site B can run
        work = self.queue.getWork({'T2_XX_SiteB': 1000, 'T2_XX_SiteAA': 1000}, {})
        self.assertEqual(len(work), 31)

    def testQueueChaining(self):
        """
        Chain WorkQueues, pull work down and verify splitting
        """
        self.assertEqual(0, len(self.globalQueue))
        # check no work in local queue
        self.assertEqual(0, len(self.localQueue.getWork({'T2_XX_SiteA': 1000}, {})))
        # Add work to top most queue
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteD"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        self.assertEqual(NBLOCKS_HICOMM, len(self.globalQueue))

        # check work isn't passed down to site without subscription
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteD': 1000}), 0)

        # put at correct site
        self.globalQueue.updateLocationInfo()

        # check work isn't passed down to the wrong agent
        work = self.localQueue.getWork({'T2_XX_SiteD': 1000}, {})  # Not in subscription
        self.assertEqual(0, len(work))
        self.assertEqual(NBLOCKS_HICOMM, len(self.globalQueue))

        # pull work down to the lowest queue
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1000}), NBLOCKS_HICOMM)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue), NBLOCKS_HICOMM)

        self.localQueue.updateLocationInfo()
        work = self.localQueue.getWork({'T2_XX_SiteA': 1000},
                                       {})
        self.assertEqual(0, len(self.localQueue))
        self.assertEqual(NBLOCKS_HICOMM, len(work))

        # check work in local and subscription made
        for x in work:
            self.assertTrue(x['SubscriptionId'] > 0)
        for x in self.localQueue.status():
            self.assertTrue(x['SubscriptionId'] > 0)
        # mark work done & check this passes upto the top level
        self.localQueue.setStatus('Done', [x.id for x in work])

    def testQueueChainingStatusUpdates(self):
        """Chain workQueues, pass work down and verify lifecycle"""

        self.assertEqual(0, len(self.globalQueue))
        self.assertEqual(0, len(self.localQueue.getWork({'T2_XX_SiteA': 1000}, {})))

        # Add work to top most queue
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteD"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        self.globalQueue.processInboundWork()
        self.assertEqual(NBLOCKS_HICOMM, len(self.globalQueue))

        # pull to local queue
        self.globalQueue.updateLocationInfo()
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1000}), NBLOCKS_HICOMM)
        syncQueues(self.localQueue)  # Tell parent local has acquired
        self.assertEqual(len(self.globalQueue.status('Acquired')), NBLOCKS_HICOMM)
        self.assertEqual(len(self.localQueue.status('Available')), NBLOCKS_HICOMM)

        # run work
        self.globalQueue.updateLocationInfo()
        work = self.localQueue.getWork({'T2_XX_SiteA': 1000},
                                       {})
        self.assertEqual(len(work), NBLOCKS_HICOMM)

        # resend info
        syncQueues(self.localQueue)
        self.assertEqual(len(self.globalQueue.status('Running')), NBLOCKS_HICOMM)
        self.assertEqual(len(self.localQueue.status('Running')), NBLOCKS_HICOMM)

        # finish work locally and propagate to global
        self.localQueue.doneWork([x.id for x in work])
        # just update the elements but not using the result
        for x in work:
            self.localQueue.backend.updateElements(x.id, PercentComplete=100, PercentSuccess=99)
        elements = self.localQueue.status('Done')
        self.assertEqual(len(elements), len(work))
        self.assertEqual([x['PercentComplete'] for x in elements],
                         [100] * len(work))
        self.assertEqual([x['PercentSuccess'] for x in elements],
                         [99] * len(work))
        syncQueues(self.localQueue, skipWMBS=True)

        elements = self.globalQueue.status('Done')
        self.assertEqual(len(elements), NBLOCKS_HICOMM)
        self.assertEqual([x['PercentComplete'] for x in elements], [100] * NBLOCKS_HICOMM)
        self.assertEqual([x['PercentSuccess'] for x in elements], [99] * NBLOCKS_HICOMM)

        self.globalQueue.performQueueCleanupActions()
        # gq still has 2 elements since elements won't be cleaned up untill requset status updated
        self.assertEqual(NBLOCKS_HICOMM, len(self.globalQueue.status()))
        elements = self.globalQueue.backend.getInboxElements('Done')
        self.assertEqual(len(elements), 1)
        self.assertEqual([x['PercentComplete'] for x in elements], [100])
        self.assertEqual([x['PercentSuccess'] for x in elements], [99])

    def testMultiTaskProduction(self):
        """
        Test Multi top level task production spec.
        multiTaskProduction spec consist 2 top level tasks each task has event size 1000 and 2000
        respectfully
        """
        # TODO: needs more rigorous test on each element per task
        # Basic production Spec
        spec = MultiTaskProductionWorkload
        for task in spec.taskIterator():
            delattr(task.steps().data.application.configuration, 'configCacheUrl')
        spec.setSpecUrl(os.path.join(self.workDir, 'multiTaskProduction.spec'))
        spec.setOwnerDetails("evansde77", "DMWM", {'dn': 'MyDN'})
        spec.save(spec.specUrl())

        specfile = spec.specUrl()
        numElements = 3
        njobs = [10] * numElements  # array of jobs per block
        total = sum(njobs)

        # Queue Work &njobs check accepted
        self.queue.queueWork(specfile)
        self.assertEqual(2, len(self.queue))

        # try to get work
        work = self.queue.getWork({'T2_XX_SiteA': 0},
                                  {})
        self.assertEqual([], work)
        # check individual task whitelists obeyed when getting work
        work = self.queue.getWork({'T2_XX_SiteA': total},
                                  {})
        self.assertEqual(len(work), 1)
        work2 = self.queue.getWork({'T2_XX_SiteB': total},
                                   {})
        self.assertEqual(len(work2), 1)
        work.extend(work2)
        self.assertEqual(len(work), 2)
        self.assertEqual(sum([x['Jobs'] for x in self.queue.status(status='Running')]),
                         total)
        # check we have all tasks and no extra/missing ones
        for task in spec.taskIterator():
            # note: order of elements in work is undefined (both inserted simultaneously)
            element = [x for x in work if x['Subscription']['workflow'].task == task.getPathName()]
            if not element:
                self.fail("Top level task %s not in wmbs" % task.getPathName())
            element = element[0]

            # check restrictions - only whitelist for now
            whitelist = element['Subscription'].getWhiteBlackList()
            whitelist = [x['site_name'] for x in whitelist if x['valid'] == 1]
            self.assertEqual(sorted(task.siteWhitelist()), sorted(whitelist))

        # no more work available
        self.assertEqual(0, len(self.queue.getWork({'T2_XX_SiteA': total, 'T2_XX_SiteB': total},
                                                   {})))
        try:
            os.unlink(specfile)
        except OSError:
            pass

    def testTeams(self):
        """
        Team behaviour
        """
        specfile = self.spec.specUrl()
        self.globalQueue.queueWork(specfile, team='The A-Team')
        self.globalQueue.processInboundWork()
        self.assertEqual(1, len(self.globalQueue))
        slots = {'T2_XX_SiteA': 1000, 'T2_XX_SiteB': 1000}

        # Can't get work for wrong team
        self.localQueue.params['Team'] = 'other'
        self.assertEqual(self.localQueue.pullWork(slots), 0)
        # and with correct team name
        self.localQueue.params['Team'] = 'The A-Team'
        self.assertEqual(self.localQueue.pullWork(slots), 1)
        syncQueues(self.localQueue)
        # when work leaves the queue in the agent it doesn't care about teams
        self.localQueue.params['Team'] = 'other'
        self.assertEqual(len(self.localQueue.getWork(slots, {})), 1)
        self.assertEqual(0, len(self.globalQueue))

    def testSplittingLargeInputs(self):
        """
        _testSplittingLargeInputs_

        Check that we can split large inputs and store the processed inputs
        in the inbox element correctly.
        """
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        inboxElement = self.globalQueue.backend.getInboxElements(elementIDs=[processingSpec.name()])
        self.assertEqual(len(inboxElement[0]['ProcessedInputs']), NBLOCKS_HICOMM)
        return

    def testGlobalBlockSplitting(self):
        """Block splitting at global level"""
        # force global queue to split work on block
        self.globalQueue.params['SplittingMapping']['DatasetBlock']['name'] = 'Block'
        self.globalQueue.params['SplittingMapping']['Block']['name'] = 'Block'
        self.globalQueue.params['SplittingMapping']['Dataset']['name'] = 'Block'

        # queue work, globally for block, pass down, report back -> complete
        totalSpec = 1
        totalBlocks = totalSpec * NBLOCKS_HICOMM
        self.assertEqual(0, len(self.globalQueue))
        for _ in range(totalSpec):
            processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB"]})
            self.globalQueue.queueWork(processingSpec.specUrl())
        self.globalQueue.processInboundWork()
        self.assertEqual(totalBlocks, len(self.globalQueue))
        # both blocks in global belong to same parent, but have different inputs
        status = self.globalQueue.status()
        self.assertEqual(status[0]['ParentQueueId'], status[1]['ParentQueueId'])
        self.assertNotEqual(status[0]['Inputs'], status[1]['Inputs'])

        # pull to local
        # location info should already be added
        # self.globalQueue.updateLocationInfo()
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1000}),
                         totalBlocks)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue.status(status='Available')),
                         totalBlocks)  # 2 in local
        # self.localQueue.updateLocationInfo()
        work = self.localQueue.getWork({'T2_XX_SiteA': 1000, 'T2_XX_SiteB': 1000},
                                       {})
        self.assertEqual(len(work), totalBlocks)
        # both refer to same wmspec
        self.assertEqual(work[0]['RequestName'], work[1]['RequestName'])
        self.localQueue.doneWork([str(x.id) for x in work])
        # elements in local deleted at end of update, only global ones left
        self.assertEqual(len(self.localQueue.status(status='Done')),
                         totalBlocks)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue.status(status='Done')),
                         totalBlocks)
        self.assertEqual(len(self.globalQueue.status(status='Done')),
                         totalBlocks)

    def testGlobalDatasetSplitting(self):
        """Dataset splitting at global level"""

        dqmHarvArgs = DQMHarvestWorkloadFactory.getTestArguments()
        dqmHarvArgs["DQMConfigCacheID"] = createConfig(dqmHarvArgs["CouchDBName"])
        factory = DQMHarvestWorkloadFactory()
        dqmWorkload = factory.factoryWorkloadConstruction('dqmTest', dqmHarvArgs)
        dqmWorkload.setSpecUrl((os.path.join(self.workDir, 'dqmTest.spec')))
        dqmWorkload.setSiteWhitelist('T2_XX_SiteA')
        dqmWorkload.setTrustLocationFlag()
        dqmWorkload.save(dqmWorkload.specUrl())

        # force global queue to split work on Dataset
        self.globalQueue.params['SplittingMapping']['DatasetBlock']['name'] = 'Dataset'
        self.globalQueue.params['SplittingMapping']['Block']['name'] = 'Dataset'
        self.globalQueue.params['SplittingMapping']['Dataset']['name'] = 'Dataset'

        # queue work, globally for block, pass down, report back -> complete
        totalSpec = 1
        self.assertEqual(0, len(self.globalQueue))
        for _ in range(totalSpec):
            self.globalQueue.queueWork(dqmWorkload.specUrl())
        # now we have elements inserted in workqueue
        self.assertEqual(len(self.globalQueue.status(status='Available')), totalSpec)
        # this should be a continuous processing of the spec, so no changes
        self.globalQueue.processInboundWork()
        self.assertEqual(totalSpec, len(self.globalQueue))
        self.assertEqual(len(self.globalQueue.status(status='Available')), totalSpec)

        # pull to local
        # self.globalQueue.updateLocationInfo()
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1000}), totalSpec)
        syncQueues(self.localQueue)
        # give a few extra seconds for work to move between local/global queues
        time.sleep(2)
        self.assertEqual(len(self.localQueue.status(status='Available')), totalSpec)
        self.assertEqual(len(self.globalQueue.status(status='Acquired')), totalSpec)
        self.localQueue.updateLocationInfo()
        work = self.localQueue.getWork({'T2_XX_SiteA': 1000}, {})
        self.assertEqual(len(work), 1)
        self.assertEqual(len(self.localQueue.status(status='Running')), totalSpec)
        self.assertEqual(len(self.globalQueue.status(status='Acquired')), totalSpec)
        self.localQueue.doneWork([str(x.id) for x in work])
        self.assertEqual(len(self.localQueue.status(status='Done')), totalSpec)
        syncQueues(self.localQueue)
        # elements are not deleted until request status is changed
        self.assertEqual(len(self.localQueue.status(status='Done')), totalSpec)
        self.assertEqual(len(self.globalQueue.status(status='Done')), totalSpec)

    def testResetWork(self):
        """Reset work in global to different child queue"""
        # TODO: This test sometimes fails - i suspect a race condition (maybe conflict in couch)
        # Cancel code needs reworking so this will hopefully be fixed then
        totalBlocks = NBLOCKS_HICOMM
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        self.globalQueue.updateLocationInfo()
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1000}),
                         totalBlocks)
        syncQueues(self.localQueue)
        work = self.localQueue.getWork({'T2_XX_SiteA': 1000, 'T2_XX_SiteB': 1000},
                                       {})
        self.assertEqual(len(work), totalBlocks)
        self.assertEqual(len(self.localQueue.status(status='Running')), totalBlocks)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.globalQueue.status(status='Running')), totalBlocks)

        # Re-assign work in global
        self.globalQueue.resetWork([x.id for x in self.globalQueue.status(status='Running')])

        # work should be canceled in local
        # TODO: Note the work in local will be orphaned but not canceled
        syncQueues(self.localQueue)
        work_at_local = [x for x in self.globalQueue.status(status='Running') \
                         if x['ChildQueueUrl'] == sanitizeURL(self.localQueue.params['QueueURL'])['url']]
        self.assertEqual(len(work_at_local), 0)

        # now 2nd queue calls and acquires work
        self.assertEqual(self.localQueue2.pullWork({'T2_XX_SiteA': 1000}),
                         totalBlocks)
        syncQueues(self.localQueue2)

        # check work in global assigned to local2
        self.assertEqual(len(self.localQueue2.status(status='Available')), totalBlocks)  # work in local2
        work_at_local2 = [x for x in self.globalQueue.status(status='Acquired')
                          if x['ChildQueueUrl'] == sanitizeURL(self.localQueue2.params['QueueURL'])['url']]
        self.assertEqual(len(work_at_local2), totalBlocks)

    def testCancelWork(self):
        """Cancel work"""
        processingSpec = self.setupReReco(assignArgs={"SiteWhitelist":["T2_XX_SiteA", "T2_XX_SiteB"]})
        self.queue.queueWork(processingSpec.specUrl())
        elements = len(self.queue)
        self.queue.updateLocationInfo()
        self.assertEqual(len(self.queue.status()), NBLOCKS_HICOMM)
        work = self.queue.getWork({'T2_XX_SiteA': 1000, 'T2_XX_SiteB': 1000}, {})
        self.assertEqual(len(self.queue), 0)
        self.assertEqual(len(self.queue.status(status='Running')), elements)
        ids = [x.id for x in work]
        self.assertEqual(len(ids), NBLOCKS_HICOMM)
        canceled = self.queue.cancelWork(ids)
        self.assertEqual(sorted(canceled), sorted(ids))
        self.assertEqual(len(self.queue.status()), NBLOCKS_HICOMM)
        self.assertEqual(len(self.queue.status(status='Running')), NBLOCKS_HICOMM)
        self.assertEqual(len(self.queue.statusInbox(status='Canceled')), 1)

        # create a new request with one fake file
        self.queue.queueWork(self.spec.specUrl())
        self.assertEqual(len(self.queue), 1)
        work = self.queue.getWork({'T2_XX_SiteA': 1000, 'T2_XX_SiteB': 1000}, {})
        self.assertEqual(len(self.queue.status(status='Running')), len(self.queue.status()))
        ids = [x.id for x in work]
        canceled = self.queue.cancelWork(WorkflowName='testProduction')
        self.assertEqual(canceled, ids)
        self.assertEqual(len(self.queue), 0)

    def testCancelWorkGlobal(self):
        """Cancel work in global queue"""
        # queue to global & pull an element to local
        processingSpec = self.setupReReco(assignArgs= {'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 1)
        syncQueues(self.localQueue)

        # cancel in global and propagate to local
        service = WorkQueueService(self.localQueue.backend.parentCouchUrlWithAuth)
        service.cancelWorkflow(processingSpec.name())
        # marked for cancel
        self.assertEqual(len(self.globalQueue.status(status='CancelRequested')), NBLOCKS_HICOMM)
        self.assertEqual(len(self.globalQueue.statusInbox(status='Acquired')), 1)

        # will cancel element left in global, one sent to local queue stays CancelRequested
        syncQueues(self.globalQueue)
        self.assertEqual(len(self.globalQueue.status(status='CancelRequested')), 1)
        self.assertEqual(len(self.globalQueue.status(status='Canceled')), NBLOCKS_HICOMM - 1)
        self.assertEqual(len(self.globalQueue.statusInbox(status='CancelRequested')), 1)
        # global parent stays CancelRequested till child queue cancels
        syncQueues(self.globalQueue)
        self.assertEqual(len(self.globalQueue.status(status='CancelRequested')), 1)
        self.assertEqual(len(self.globalQueue.status(status='Canceled')), NBLOCKS_HICOMM - 1)
        self.assertEqual(len(self.globalQueue.statusInbox(status='CancelRequested')), 1)

        # during sync local queue will synced with globalQueue but not gets deleted until workflow finished
        syncQueues(self.localQueue)

        self.assertEqual(len(self.localQueue.statusInbox(status='Canceled')), 1)  # inbox is synced
        self.assertEqual(len(self.globalQueue.status(status='Canceled')), NBLOCKS_HICOMM)
        self.assertEqual(len(self.globalQueue.statusInbox(status='CancelRequested')), 1)
        syncQueues(self.globalQueue)
        self.assertEqual(len(self.globalQueue.status(status='Canceled')), NBLOCKS_HICOMM)
        self.assertEqual(len(self.globalQueue.statusInbox(status='Canceled')), 1)
        syncQueues(self.localQueue)
        # local cancelded
        # self.assertEqual(len(self.localQueue.status(status='Canceled')), 1)
        # clear global
        self.globalQueue.deleteWorkflows(processingSpec.name())
        self.assertEqual(len(self.globalQueue.statusInbox()), 0)

        ### check cancel of work negotiating in agent works
        self.globalQueue.queueWork(self.whitelistSpec.specUrl())
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteB': 1}), 1)
        self.localQueue.backend.forceQueueSync()
        time.sleep(2)
        self.assertEqual(len(self.localQueue.statusInbox(status='Negotiating')), 1)

        # now cancel
        service.cancelWorkflow(self.whitelistSpec.name())
        syncQueues(self.globalQueue)
        self.localQueue.backend.forceQueueSync()  # pull in cancelation
        time.sleep(2)
        self.assertEqual(len(self.globalQueue.status(status='Canceled')), 2 * NBLOCKS_HICOMM - 1)
        self.assertEqual(len(self.localQueue.statusInbox(status='CancelRequested')), 1)
        syncQueues(self.localQueue, skipWMBS=True)
        self.assertEqual(len(self.localQueue.statusInbox(status='Canceled')), 2)
        syncQueues(self.localQueue)
        syncQueues(self.globalQueue)

        self.assertEqual(len(self.localQueue.statusInbox(WorkflowName=self.whitelistSpec.name())), 1)
        self.assertEqual(len(self.globalQueue.status(WorkflowName=self.whitelistSpec.name())), NBLOCKS_HICOMM)
        self.assertEqual(len(self.globalQueue.statusInbox(status='Canceled')), 1)
        # clear global
        self.globalQueue.deleteWorkflows(self.whitelistSpec.name())
        self.assertEqual(len(self.globalQueue.statusInbox()), 0)

    def testInvalidSpecs(self):
        """Complain on invalid WMSpecs"""
        # request != workflow name
        self.assertRaises(WorkQueueWMSpecError, self.queue.queueWork,
                          self.setupReReco().specUrl(),
                          request='fail_this')

        # invalid white list
        mcspec = stepchainWorkload('testProductionInvalid', self.mcArgs)
        getFirstTask(mcspec).setSiteWhitelist('ThisIsInvalid')
        mcspec.setSpecUrl(os.path.join(self.workDir, 'testProductionInvalid.spec'))
        mcspec.save(mcspec.specUrl())
        self.assertRaises(WorkQueueWMSpecError, self.queue.queueWork, mcspec.specUrl())
        getFirstTask(mcspec).setSiteWhitelist(['T2_XX_SiteB'])
        self.queue.deleteWorkflows(mcspec.name())

        # 0 events
        getFirstTask(mcspec).addProduction(totalEvents=0)
        getFirstTask(mcspec).setSiteWhitelist(['T2_XX_SiteB'])
        mcspec.save(mcspec.specUrl())
        self.assertRaises(WorkQueueNoWorkError, self.queue.queueWork, mcspec.specUrl())

        # no dataset
        processingSpec = rerecoWorkload('testProcessingInvalid', self.rerecoArgs)
        getFirstTask(processingSpec).setSiteWhitelist(['T2_XX_SiteB'])
        processingSpec.setSpecUrl(os.path.join(self.workDir,
                                               'testProcessingInvalid.spec'))
        processingSpec.save(processingSpec.specUrl())
        getFirstTask(processingSpec).data.input.dataset = None
        processingSpec.save(processingSpec.specUrl())
        self.assertRaises(WorkQueueWMSpecError, self.queue.queueWork, processingSpec.specUrl())

        # invalid dbs url
        processingSpec = rerecoWorkload('testProcessingInvalid', self.rerecoArgs)
        getFirstTask(processingSpec).setSiteWhitelist(['T2_XX_SiteB'])
        processingSpec.setSpecUrl(os.path.join(self.workDir,
                                               'testProcessingInvalid.spec'))
        getFirstTask(processingSpec).data.input.dataset.dbsurl = 'wrongprot://dbs.example.com'
        processingSpec.save(processingSpec.specUrl())
        self.assertRaises(WorkQueueWMSpecError, self.queue.queueWork, processingSpec.specUrl())
        self.queue.deleteWorkflows(processingSpec.name())

        # invalid dataset name
        processingSpec = rerecoWorkload('testProcessingInvalid', self.rerecoArgs)
        getFirstTask(processingSpec).setSiteWhitelist(['T2_XX_SiteB'])
        processingSpec.setSpecUrl(os.path.join(self.workDir,
                                               'testProcessingInvalid.spec'))
        getFirstTask(processingSpec).data.input.dataset.name = '/MinimumBias/FAKE-Filter-v1/RECO'
        processingSpec.save(processingSpec.specUrl())
        self.assertRaises(DBSReaderError, self.queue.queueWork, processingSpec.specUrl())
        self.queue.deleteWorkflows(processingSpec.name())

        # Cant have a slash in primary ds name - validation should fail
        getFirstTask(processingSpec).data.input.dataset.primary = 'a/b'
        processingSpec.save(processingSpec.specUrl())
        self.assertRaises(DBSReaderError, self.queue.queueWork, processingSpec.specUrl())
        self.queue.deleteWorkflows(processingSpec.name())

        # dataset splitting with invalid run whitelist
        processingSpec = rerecoWorkload('testProcessingInvalid', self.rerecoArgs)
        getFirstTask(processingSpec).setSiteWhitelist(['T2_XX_SiteB'])
        processingSpec.setSpecUrl(os.path.join(self.workDir,
                                               'testProcessingInvalid.spec'))
        processingSpec.setStartPolicy('Dataset')
        processingSpec.setRunWhitelist([666])  # not in this dataset
        processingSpec.save(processingSpec.specUrl())
        self.assertRaises(DBSReaderError, self.queue.queueWork, processingSpec.specUrl())
        self.queue.deleteWorkflows(processingSpec.name())

        # block splitting with invalid run whitelist
        processingSpec = rerecoWorkload('testProcessingInvalid', self.rerecoArgs)
        getFirstTask(processingSpec).setSiteWhitelist(['T2_XX_SiteB'])
        processingSpec.setSpecUrl(os.path.join(self.workDir,
                                               'testProcessingInvalid.spec'))
        processingSpec.setStartPolicy('Block')
        processingSpec.setRunWhitelist([666])  # not in this dataset
        processingSpec.save(processingSpec.specUrl())
        self.assertRaises(DBSReaderError, self.queue.queueWork, processingSpec.specUrl())
        self.queue.deleteWorkflows(processingSpec.name())

    def testIgnoreDuplicates(self):
        """Ignore duplicate work"""
        specfile = self.spec.specUrl()
        self.globalQueue.queueWork(specfile)
        self.assertEqual(1, len(self.globalQueue))

        # queue work again
        self.globalQueue.queueWork(specfile)
        self.assertEqual(1, len(self.globalQueue))

    def testConflicts(self):
        """Resolve conflicts between global & local queue"""
        self.globalQueue.queueWork(self.spec.specUrl())
        self.localQueue.pullWork({'T2_XX_SiteA': TOTAL_EVENTS})
        self.localQueue.getWork({'T2_XX_SiteA': TOTAL_EVENTS},
                                {})
        syncQueues(self.localQueue)
        global_ids = [x.id for x in self.globalQueue.status()]
        self.localQueue.backend.updateInboxElements(*global_ids, Status='Done', PercentComplete=69)
        self.globalQueue.backend.updateElements(*global_ids, Status='Canceled')
        self.localQueue.backend.forceQueueSync()
        time.sleep(2)
        self.assertForConflicts()

    @retry(AssertionError, tries=3, delay=10)
    def assertForConflicts(self):
        """
        Make the assertions in a separate function so we can use retry
        """

        global_ids = [x.id for x in self.globalQueue.status()]

        self.localQueue.backend.fixConflicts()
        self.localQueue.backend.forceQueueSync()
        time.sleep(2)
        self.assertEqual([x['Status'] for x in self.globalQueue.status(elementIDs=global_ids)],
                         ['Canceled'])
        self.assertEqual([x['PercentComplete'] for x in self.globalQueue.status(elementIDs=global_ids)],
                         [69])
        self.assertEqual([x for x in self.localQueue.statusInbox()],
                         [x for x in self.globalQueue.status()])

    def testDeleteWork(self):
        """Delete finished work
        TODO: do emulate the reqmgr2 and change the status of request
        so actually request gets deleted when performCleanupAction is run.
        """
        self.globalQueue.queueWork(self.spec.specUrl())
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': TOTAL_EVENTS}), 1)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue.getWork({'T2_XX_SiteA': TOTAL_EVENTS},
                                                     {})), 1)
        syncQueues(self.localQueue)
        self.localQueue.doneWork(WorkflowName=self.spec.name())
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue.status(WorkflowName=self.spec.name())),
                         1)  # not deleted until request status is updated
        self.assertEqual('Done',
                         self.globalQueue.status(WorkflowName=self.spec.name())[0]['Status'])
        self.globalQueue.performQueueCleanupActions()
        self.assertEqual('Done',
                         self.globalQueue.statusInbox(WorkflowName=self.spec.name())[0]['Status'])
        self.assertEqual(len(self.globalQueue.status(WorkflowName=self.spec.name())),
                         1)  # not deleted until request status is update
        self.globalQueue.deleteWorkflows(self.spec.name())
        self.assertEqual(len(self.globalQueue.statusInbox(WorkflowName=self.spec.name())),
                         0)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue.statusInbox(WorkflowName=self.spec.name())),
                         1)  # not deleted until request status is update

    def testResubmissionWorkflow(self):
        """Test workflow resubmission via ACDC"""
        acdcCouchDB = "workqueue_t_acdc"
        self.testInit.setupCouch(acdcCouchDB, "GroupUser", "ACDC")

        spec = self.createResubmitSpec(self.testInit.couchUrl,
                                       acdcCouchDB)
        spec.setSpecUrl(os.path.join(self.workDir, 'resubmissionWorkflow.spec'))
        spec.setSiteWhitelist('T1_US_FNAL')
        spec.save(spec.specUrl())
        self.localQueue.params['Team'] = 'cmsdataops'
        self.globalQueue.queueWork(spec.specUrl(), "Resubmit_TestWorkload", team="cmsdataops")
        self.assertEqual(self.localQueue.pullWork({"T1_US_FNAL": 100}), 1)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue.getWork({"T1_US_FNAL": 100}, {})), 1)

    def testResubmissionWithParentsWorkflow(self):
        """Test workflow resubmission with parentage via ACDC"""
        acdcCouchDB = "workqueue_t_acdc"
        self.testInit.setupCouch(acdcCouchDB, "GroupUser", "ACDC")

        spec = self.createResubmitSpec(self.testInit.couchUrl,
                                       acdcCouchDB, parentage=True)
        spec.setSpecUrl(os.path.join(self.workDir, 'resubmissionWorkflow.spec'))
        spec.setSiteWhitelist('T1_US_FNAL')
        spec.save(spec.specUrl())
        self.localQueue.params['Team'] = 'cmsdataops'
        self.globalQueue.queueWork(spec.specUrl(), "Resubmit_TestWorkload", team="cmsdataops")
        self.localQueue.pullWork({"T1_US_FNAL": 100})
        syncQueues(self.localQueue)
        self.localQueue.getWork({"T1_US_FNAL": 100}, {})

    def testResubmissionWorkflowSiteWhitelistLocations(self):
        """ Test an ACDC workflow where we use the site whitelist as locations"""
        acdcCouchDB = "workqueue_t_acdc"
        self.testInit.setupCouch(acdcCouchDB, "GroupUser", "ACDC")

        spec = self.createResubmitSpec(self.testInit.couchUrl,
                                       acdcCouchDB)
        spec.setSpecUrl(os.path.join(self.workDir, 'resubmissionWorkflow.spec'))
        spec.setSiteWhitelist('T1_US_FNAL')
        spec.setTrustLocationFlag(inputFlag=True, pileupFlag=False)
        spec.save(spec.specUrl())
        self.localQueue.params['Team'] = 'cmsdataops'
        self.globalQueue.queueWork(spec.specUrl(), "Resubmit_TestWorkload", team="cmsdataops")
        self.assertEqual(self.localQueue.pullWork({"T1_UK_RAL": 100}), 0)
        self.assertEqual(self.localQueue.pullWork({"T1_US_FNAL": 100}), 1)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue.getWork({"T1_US_FNAL": 100}, {})), 1)

    def testThrottling(self):
        """Pull work only if all previous work processed in child"""
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA"]})

        self.globalQueue.queueWork(processingSpec.specUrl())
        self.assertEqual(NBLOCKS_HICOMM, len(self.globalQueue))
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 1)
        # further pull will fail till we replicate to child
        # hopefully couch replication wont happen till we manually sync
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 0)
        self.assertEqual(NBLOCKS_HICOMM - 1, len(self.globalQueue))
        self.assertEqual(0, len(self.localQueue))
        syncQueues(self.localQueue)
        self.assertEqual(1, len(self.localQueue))
        # pull works again
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 1)

    def testSitesFromResourceControl(self):
        """Test sites from resource control"""
        # Most tests pull work for specific sites (to give us control)
        # In reality site list will come from resource control so test
        # that here (just a simple check that we can get sites from rc)
        self.globalQueue.queueWork(self.spec.specUrl())
        self.assertEqual(self.localQueue.pullWork(), 1)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue.status()), 1)

    def testParentProcessing(self):
        """
        Enqueue and get work for a processing WMSpec.
        """
        parentProcSpec = self.setupParentProcSpec(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB"]})
        specfile = parentProcSpec.specUrl()

        # Queue Work & check accepted
        self.queue.queueWork(specfile)
        self.queue.processInboundWork()
        self.assertEqual(NBLOCKS_COSMIC, len(self.queue))

        self.queue.updateLocationInfo()
        # No resources
        work = self.queue.getWork({}, {})
        self.assertEqual(len(work), 0)
        work = self.queue.getWork({'T2_XX_SiteA': 0,
                                   'T2_XX_SiteB': 0}, {})
        self.assertEqual(len(work), 0)

        # Get 1 work element when any resources free
        work = self.queue.getWork({'T2_XX_SiteB': 1}, {})
        self.assertEqual(len(work), 1)
        processedFiles = work[0]["NumOfFilesAdded"]

        # claim remaining work
        work = self.queue.getWork({'T2_XX_SiteA': TOTAL_EVENTS, 'T2_XX_SiteB': TOTAL_EVENTS}, {})
        self.assertEqual(len(work), NBLOCKS_COSMIC - 1)
        for element in work:
            processedFiles += element["NumOfFilesAdded"]
        self.assertEqual(processedFiles, NFILES_COSMIC + NFILES_COSMICRAW)

        # no more work available
        self.assertEqual(0, len(self.queue.getWork({'T2_XX_SiteA': 1000}, {})))

    def testWMBSInjectionStatus(self):
        syncQueues(self.localQueue)
        self.globalQueue.queueWork(self.spec.specUrl())
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        # test globalqueue status (no parent queue case)
        self.assertEqual(self.globalQueue.getWMBSInjectionStatus(),
                         [{'testProcessing': False}, {'testProduction': False}])
        self.assertEqual(self.globalQueue.getWMBSInjectionStatus(self.spec.name()),
                         False)
        # Amount of work varies, just make sure it's positive
        self.assertGreater(self.localQueue.pullWork(), 0)
        # test local queue status with parents (globalQueue is not synced yet
        self.assertEqual(self.localQueue.getWMBSInjectionStatus(),
                         [{'testProcessing': False}, {'testProduction': False}])
        self.assertEqual(self.localQueue.getWMBSInjectionStatus(self.spec.name()),
                         False)
        self.localQueue.processInboundWork()
        self.localQueue.updateLocationInfo()
        self.localQueue.getWork({'T2_XX_SiteA': 1000},
                                {})
        self.assertEqual(self.localQueue.getWMBSInjectionStatus(),
                         [{'testProcessing': False}, {'testProduction': False}])
        self.assertEqual(self.localQueue.getWMBSInjectionStatus(self.spec.name()),
                         False)

        # update parents status but is still running open since it is the default
        self.localQueue.performQueueCleanupActions()
        # NOTE: these 2 assertions below are different when running from local docker container
        self.assertEqual(self.localQueue.getWMBSInjectionStatus(),
                         [{'testProcessing': False}, {'testProduction': False}])
        self.assertEqual(self.localQueue.getWMBSInjectionStatus(self.spec.name()), False)

        # close the global inbox elements, they won't be split anymore
        #self.globalQueue.closeWork(['testProcessing', 'testProduction'])
        self.globalQueue.closeWork()
        self.localQueue.getWMBSInjectionStatus()
        time.sleep(1)
        # There are too many jobs to pull down for testProcessing still has element not in WMBS
        self.assertEqual(self.localQueue.getWMBSInjectionStatus(),
                         [{'testProcessing': False}, {'testProduction': True}])
        self.assertEqual(self.localQueue.getWMBSInjectionStatus(self.spec.name()),
                         True)

        # test not existing workflow
        self.assertRaises(WorkQueueNoMatchingElements,
                          self.localQueue.getWMBSInjectionStatus,
                          "NotExistWorkflow")

    def testEndPolicyNegotiating(self):
        """Test end policy processing of request before splitting"""
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA"]})
        work = self.globalQueue.queueWork(processingSpec.specUrl())
        self.assertEqual(work, NBLOCKS_HICOMM)
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 1)
        self.localQueue.backend.pullFromParent()  # pull work into inbox (Negotiating state)
        self.localQueue.processInboundWork()
        syncQueues(self.localQueue)
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 1)
        # should print message but not raise an error
        self.localQueue.performQueueCleanupActions(skipWMBS=True)
        self.localQueue.backend.pullFromParent(continuous=False)
        self.assertEqual(len(self.localQueue.statusInbox(Status='Negotiating')), 1)
        self.assertEqual(len(self.localQueue), 1)

    def testSiteStatus(self):
        """Check that we only pull work on sites in Normal status"""
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        self.globalQueue.queueWork(self.spec.specUrl())
        # acquire 1 element of a wf and then mark site as draining.
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 1)
        syncQueues(self.localQueue)
        existing_wf = [x['RequestName'] for x in self.localQueue.statusInbox()]
        self.assertEqual(1, len(existing_wf))
        existing_wf = existing_wf[0]
        bossAirConfig = Configuration()
        bossAirConfig.section_("BossAir")
        bossAirConfig.BossAir.pluginDir = "WMCore.BossAir.Plugins"
        bossAirConfig.BossAir.pluginNames = []
        bossAirConfig.section_("Agent")
        bossAirConfig.Agent.agentName = "TestAgent"
        bossAirConfig.section_("JobStateMachine")
        bossAirConfig.JobStateMachine.couchurl = os.environ["COUCHURL"]
        bossAirConfig.JobStateMachine.couchDBName = "testcouchdb"
        rc = ResourceControl(bossAirConfig)
        rc.changeSiteState('T2_XX_SiteA', 'Draining')
        rc.changeSiteState('T2_XX_SiteB', 'Draining')
        # pull more work, no work should be acquired
        self.localQueue.pullWork()
        syncQueues(self.localQueue)
        for x in self.localQueue.statusInbox():
            if x['RequestName'] != existing_wf:
                self.fail('Got new wf %s for draining site' % x['RequestName'])
        # wmbs injection for draining sites continues to work
        self.assertTrue(self.localQueue.getWork({'T2_XX_SiteA': 10},
                                                {}))
        # re-enable site and get remainder of work
        rc.changeSiteState('T2_XX_SiteA', 'Normal')
        self.assertTrue(self.localQueue.pullWork())
        syncQueues(self.localQueue)
        self.assertTrue(self.localQueue.getWork({'T2_XX_SiteA': 100},
                                                {}))

    def test0eventBlock(self):
        """0 event blocks should be processed as usual"""
        # use event splitting and 0 events so we get 0 jobs - verify this doesn't cause any problems
        # FIXME: This does not work currently because we don't actually have 0 event blocks.

        Globals.GlobalParams.setNumOfEventsPerFile(0)
        syncQueues(self.localQueue)

        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA"]})
        processingSpec.setStartPolicy('Block', SliceType='NumberOfEvents')
        processingSpec.save(processingSpec.specUrl())
        self.globalQueue.queueWork(processingSpec.specUrl())
        # pulls one block from global queue to local queue inbox
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 1)
        time.sleep(1)
        # now we should have 1 element in the local inbox and 0 in local queue
        self.assertEqual(len(self.localQueue.statusInbox()), 1)
        self.assertEqual(len(self.localQueue.status()), 0)

        self.localQueue.processInboundWork()
        time.sleep(1)
        # now acquire the local inbox element (create it in the local queue)
        self.assertEqual(len(self.localQueue.getWork({'T2_XX_SiteA': 1}, {})), 1)
        for element in self.localQueue.status():
            # check files added and subscription made
            self.assertEqual(element['NumOfFilesAdded'], 1)
            self.assertTrue(element['SubscriptionId'] >= 0)
            self.assertEqual(element['Jobs'], 1)

        # complete workflow
        self.localQueue.performQueueCleanupActions(skipWMBS=True)
        self.localQueue.doneWork([str(x.id) for x in self.localQueue.status()])
        self.assertEqual(len(self.localQueue.status(status='Done')), 1)
        syncQueues(self.localQueue)
        self.assertEqual(len(self.globalQueue.status(status='Done')), 1)

    def testProcessingWithContinuousSplitting(self):
        """Test the open request handling in the WorkQueue"""
        # Put normal work in
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB", "T2_XX_SiteC"]})
        specfile = processingSpec.specUrl()

        # Queue work with initial block count
        logging.info("Queuing work for spec name: %s", processingSpec.name())
        self.assertEqual(NBLOCKS_HICOMM, self.globalQueue.queueWork(specfile))
        self.assertEqual(NBLOCKS_HICOMM, len(self.globalQueue))

        # Try adding work, no change in blocks available. No work should be added
        logging.info("Adding work - already added - for spec name: %s", processingSpec.name())
        workInbox = self.globalQueue.backend.getInboxElements(WorkflowName=processingSpec.name(),
                                                              loadSpec=True)
        self.assertEqual(0, self.globalQueue.addWork(workInbox[0]))
        self.assertEqual(NBLOCKS_HICOMM, len(self.globalQueue))

        # Now pull work from the global to the local queue
        logging.info("Pulling 1 workqueue element from the parent queue")
        self.localQueue.pullWork({'T2_XX_SiteA': 1})
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue), 1)
        self.assertEqual(len(self.globalQueue), NBLOCKS_HICOMM - 1)

        # This time pull work from the local queue into WMBS
        logging.info("Getting 1 workqueue element from the local queue")
        dummyWork = self.localQueue.getWork({'T2_XX_SiteA': 1000},
                                            {})
        syncQueues(self.localQueue)
        syncQueues(self.globalQueue)

        workInbox = self.globalQueue.backend.getInboxElements(status="Running", loadSpec=True)
        self.assertEqual(0, self.globalQueue.addWork(workInbox[0]))

        self.assertEqual(NBLOCKS_HICOMM - 1, len(self.globalQueue))
        self.assertEqual(len(self.globalQueue.backend.getInboxElements(status="Running")), 1)

        # Now pull the new work to the local queue
        self.localQueue.pullWork({'T2_XX_SiteB': 1000, 'T2_XX_SiteC': 1000})
        syncQueues(self.localQueue)
        self.assertEqual(len(self.localQueue), 30)
        self.assertEqual(len(self.globalQueue), NBLOCKS_HICOMM - 30 - 1)

        workInbox = self.globalQueue.backend.getInboxElements(WorkflowName=processingSpec.name(), loadSpec=True)
        self.assertEqual(0, self.globalQueue.addWork(workInbox[0]))

        return

    def testProcessingWithPileup(self):
        """Test a full WorkQueue cycle in a request with pileup datasets"""
        specfile = self.processingPileupSpec.specUrl()
        # Queue work with initial block count
        self.assertEqual(NBLOCKS_HICOMM, self.globalQueue.queueWork(specfile))
        self.assertEqual(NBLOCKS_HICOMM, len(self.globalQueue))

        # All blocks are in Site A, B, and C, but the pileup is only at C.
        # We should not be able to pull all the work.
        self.assertGreaterEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1,
                                                          'T2_XX_SiteB': 1,
                                                          'T2_XX_SiteC': 1}), 3)
        # The Rucio emulator will move the pileup blocks to site A
        self.globalQueue.updateLocationInfo()
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteB': 1,
                                                   'T2_XX_SiteC': 1}), 0)

        # Now try with just site A (no work)
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': 1}), 0)
        syncQueues(self.localQueue)
        self.assertGreaterEqual(len(self.localQueue), 3)
        self.assertEqual(len(self.globalQueue), NBLOCKS_HICOMM - 3)

        # Pull it to WMBS, first try with an impossible site
        # The pileup was split again in the local queue so site A is not there
        self.assertGreaterEqual(len(self.localQueue.getWork({'T2_XX_SiteA': 1,
                                                             'T2_XX_SiteB': 1,
                                                             'T2_XX_SiteC': 1}, {})), 2)
        Globals.moveBlock({'%s#1' % PILEUP_DATASET: ['T2_XX_SiteA', 'T2_XX_SiteC'],
                           '%s#2' % PILEUP_DATASET: ['T2_XX_SiteA', 'T2_XX_SiteC']})
        self.localQueue.updateLocationInfo()

        #FIXME: this test gives different results in jenkins and in private tests
        self.assertGreaterEqual(len(self.localQueue.getWork({'T2_XX_SiteA': 1}, {})), 0)

        self.assertGreaterEqual(len(self.localQueue.status()), 3)

    def testPileupOnProduction(self):
        """Test that we can split properly a Production workflow with pileup"""
        specfile = self.productionPileupSpec.specUrl()

        # Sanity check on queueWork only
        self.assertEqual(1, self.globalQueue.queueWork(specfile))
        self.assertEqual(1, len(self.globalQueue))
        self.assertEqual(len(self.globalQueue.backend.getActivePileupData()), 1)
        self.assertNotEqual(self.globalQueue.backend.getActivePileupData()[0]['dbs_url'], None)

    def testPrioritiesWorkPolling(self):
        """Test how the priorities and current jobs in the queue affect the workqueue behavior
           for acquiring and injecting work"""
        # Queue a low prio workflow and a high prio workflow
        highPrioReReco = self.setupHighPrioReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA"]})
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        self.globalQueue.queueWork(highPrioReReco.specUrl())

        # Pull all into local queue
        self.assertEqual(self.localQueue.pullWork({'T2_XX_SiteA': TOTAL_EVENTS}), 2 * NBLOCKS_HICOMM)
        syncQueues(self.localQueue)

        # Try pulling work into WMBS when "there is" a job of higher priority than the high prio workflow
        self.assertEqual(len(self.localQueue.getWork({'T2_XX_SiteA': 1},
                                                     {'T2_XX_SiteA': {highPrioReReco.priority() + 1: 1}})),
                         0)

        # Allow one more job slot
        self.assertEqual(len(self.localQueue.getWork({'T2_XX_SiteA': 2},
                                                     {'T2_XX_SiteA': {highPrioReReco.priority() + 1: 1}})),
                         1)

        # Allow 1 slot more and many slots occupied by low prio jobs
        self.assertEqual(len(self.localQueue.getWork({'T2_XX_SiteA': 2},
                                                     {'T2_XX_SiteA': {1: 50}})),
                         1)
        self.assertEqual(len(self.localQueue.backend.getElements(WorkflowName=highPrioReReco.name())),
                         NBLOCKS_HICOMM)

    def testMonitorWorkQueue(self):
        """
        Test several WorkQueue couch queries to monitor amount of work in the system
        """
        # Run some bootstrap, same code as in the test above...
        highPrioReReco = self.setupHighPrioReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB"]})
        processingSpec = self.setupReReco(assignArgs={'SiteWhitelist': ["T2_XX_SiteA", "T2_XX_SiteB"]})
        self.globalQueue.queueWork(processingSpec.specUrl())
        self.globalQueue.queueWork(highPrioReReco.specUrl())

        initialStatus = ['Available', 'Negotiating', 'Acquired']
        metrics = self.globalQueue.monitorWorkQueue(status=initialStatus)
        time.sleep(1)  # HACKY: query again to get the up-to-date views
        metrics = self.globalQueue.monitorWorkQueue(status=initialStatus)

        expectedMetrics = ('workByStatus', 'workByStatusAndPriority', 'workByAgentAndStatus',
                           'workByAgentAndPriority', 'uniqueJobsPerSiteAAA', 'possibleJobsPerSiteAAA',
                           'uniqueJobsPerSite', 'possibleJobsPerSite', 'total_query_time')
        self.assertItemsEqual(list(metrics), expectedMetrics)

        self.assertItemsEqual(list(metrics['workByStatus']), STATES)
        self.assertEqual(metrics['workByStatus']['Available']['sum_jobs'], 678)
        self.assertEqual(metrics['workByStatus']['Acquired'], {})

        self.assertItemsEqual(list(metrics['workByStatusAndPriority']), STATES)
        prios = [item['priority'] for item in metrics['workByStatusAndPriority']['Available']]
        self.assertItemsEqual(prios, [8000, 999998])
        self.assertEqual(metrics['workByStatusAndPriority']['Acquired'], [])

        self.assertEqual(len(metrics['workByAgentAndStatus']), 1)
        self.assertEqual(metrics['workByAgentAndStatus'][0]['agent_name'], 'AgentNotDefined')
        self.assertEqual(metrics['workByAgentAndStatus'][0]['status'], 'Available')

        self.assertEqual(len(metrics['workByAgentAndPriority']), 2)
        self.assertEqual(metrics['workByAgentAndPriority'][0]['agent_name'], 'AgentNotDefined')
        self.assertEqual([item['priority'] for item in metrics['workByAgentAndPriority']], [8000, 999998])

        for met in ('uniqueJobsPerSiteAAA', 'possibleJobsPerSiteAAA', 'uniqueJobsPerSite', 'possibleJobsPerSite'):
            self.assertItemsEqual(list(metrics[met]), initialStatus)
            self.assertEqual(len(metrics[met]['Available']), 2)
            self.assertEqual(len(metrics[met]['Acquired']), 0)
            self.assertItemsEqual(list(metrics[met]['Available']), ['T2_XX_SiteA', 'T2_XX_SiteB'])

        self.assertTrue(metrics['total_query_time'] >= 0)

        # Pull all into local queue (get them into Acquired status)
        self.localQueue.pullWork({'T2_XX_SiteA': 500})
        syncQueues(self.localQueue)
        metrics = self.globalQueue.monitorWorkQueue(status=initialStatus)
        time.sleep(1)  # HACKY: query again to get the up-to-date views
        metrics = self.globalQueue.monitorWorkQueue(status=initialStatus)

        self.assertTrue(metrics['workByStatus']['Available']['sum_jobs'] < 200)
        self.assertTrue(metrics['workByStatus']['Acquired']['sum_jobs'] >= 500)

        self.assertEqual(len(metrics['workByStatusAndPriority']['Available']), 1)
        self.assertEqual(len(metrics['workByStatusAndPriority']['Acquired']), 2)
        self.assertEqual(metrics['workByStatusAndPriority']['Available'][0]['priority'], 8000)
        prios = [item['priority'] for item in metrics['workByStatusAndPriority']['Acquired']]
        self.assertItemsEqual(prios, [8000, 999998])

        self.assertEqual(len(metrics['workByAgentAndStatus']), 2)
        for elem in metrics['workByAgentAndStatus']:
            if elem['status'] == 'Available':
                self.assertEqual(elem['agent_name'], 'AgentNotDefined')
            else:  # in Acquired
                self.assertTrue(elem['agent_name'] != 'AgentNotDefined')

        self.assertEqual(len(metrics['workByAgentAndPriority']), 3)
        prios = []
        for item in metrics['workByAgentAndPriority']:
            if item['agent_name'] != 'AgentNotDefined':
                prios.append(item['priority'])
        self.assertItemsEqual(prios, [8000, 999998])

        for met in ('uniqueJobsPerSiteAAA', 'possibleJobsPerSiteAAA', 'uniqueJobsPerSite', 'possibleJobsPerSite'):
            self.assertItemsEqual(list(metrics[met]), initialStatus)
            self.assertEqual(len(metrics[met]['Available']), 2)
            self.assertEqual(len(metrics[met]['Acquired']), 2)
            self.assertItemsEqual(list(metrics[met]['Available']), ['T2_XX_SiteA', 'T2_XX_SiteB'])
            self.assertItemsEqual(list(metrics[met]['Acquired']), ['T2_XX_SiteA', 'T2_XX_SiteB'])


if __name__ == "__main__":
    unittest.main()



"""
_InsertComponent_

MySQL implementation of UpdateWorker
"""

__all__ = []



import time
from WMCore.Database.DBFormatter import DBFormatter

class UpdateWorkerError(DBFormatter):

    sql = """UPDATE wm_workers
              SET last_error = :last_error,
                  state = :state, error_message = :error_message
              WHERE component_id = (SELECT id FROM wm_components WHERE name = :component_name)
                   AND name = :worker_name"""

    def execute(self, componentName, workerName, errorMessage,
                conn = None, transaction = False):
        binds = {"component_name": componentName,
                 "worker_name": workerName,
                 "last_error": int(time.time()),
                 "state": "Error",
                 "error_message": errorMessage[:1000]}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
File       : SiteListPoller
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: module to update of site lists within a WMAgent
"""

# system modules
import logging
import threading
from pprint import pformat
# WMCore modules
from Utils.Timers import timeFunction
from WMCore.DAOFactory import DAOFactory
from WMCore.Lexicon import sanitizeURL
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue
from WMCore.Services.WMStatsServer.WMStatsServer import WMStatsServer
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread


class SiteListPoller(BaseWorkerThread):
    def __init__(self, config):
        """
        Initialize SiteListPoller object
        :param config: a Configuration object with the component configuration
        """
        BaseWorkerThread.__init__(self)
        myThread = threading.currentThread()
        self.logger = myThread.logger

        # get wmstats parameters
        self.wmstatsUrl = getattr(config.WorkflowUpdater, "wmstatsUrl")
        self.wmstatsSrv = WMStatsServer(self.wmstatsUrl, logger=self.logger)
        self.reqStates = getattr(config.WorkflowUpdater, "states", ['running-open', 'acquired'])
        self.wqeStates = getattr(config.WorkflowUpdater, "wqeStates", ['Available'])

        # provide access to WMBS in local WMAgent
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        # DB function to retrieve active workflows
        self.listActiveWflows = self.daoFactory(classname="Workflow.GetUnfinishedWorkflows")

        # local WorkQueue service
        self.localCouchUrl = config.WorkQueueManager.couchurl
        self.localWQ = WorkQueue(self.localCouchUrl,
                                 config.WorkQueueManager.dbname)

    def getActiveWorkflows(self):
        """
        Provide list of active requests within WMAgent
        :return: list of workflow names
        """
        self.logger.info("Fetching active workflows in the agent")
        # get list of active workflows in WMAgent
        wflowSpecs = self.listActiveWflows.execute()

        return [wflowSpec['name'] for wflowSpec in wflowSpecs]

    def findUnacquiredWorkflows(self):
        """
        Using a non-customized CouchDB view, find out workflows with elements
        in the statuses that we would like to update (e.g. Available).
        This is required because not all workflows might have been acquired by WMBS.
        :return: a flat list of workflow names
        """
        self.logger.info("Finding not yet active workflows in local workqueue")
        response = []
        # get list of active workflows in WMAgent by looking into local elements in Available status
        summary = self.localWQ.getElementsCountAndJobsByWorkflow()
        for wflowName, innerDict in summary.items():
            for status in innerDict:
                if status in self.wqeStates:
                    response.append(wflowName)
        return response

    def wmstatsDict(self, requests):
        """
        Return list of requests specs from WMStats for provided list of request names
        :param requests: list of workflow requests names
        :return: dict of workflow records obtained from wmstats server:
        {"wflow": {"SiteWhitelist":[], "SiteBlacklist": []}, ...}
        """
        # get list of workflows from wmstats
        outputMask = ['SiteWhitelist', 'SiteBlacklist']
        wdict = {}
        for state in self.reqStates:
            inputConditions = {"RequestStatus": state}
            self.logger.info("Fetch site info from WMStats for condition: %s and mask %s", inputConditions, outputMask)
            data = self.wmstatsSrv.getFilteredActiveData(inputConditions, outputMask)
            self.logger.info("Found %d workflows in WMStats with status %s", len(data), state)
            self.logger.debug("Data from wmstats for status %s: %s", state, pformat(data))
            for rdict in data:
                # rdict here has the following structure:
                # {"RequestName": "bla", "SiteWhitelist":[], "SiteBlacklist": []}
                wflow = rdict.pop('RequestName')
                # check that our workflow is in our requests list
                if wflow in requests:
                    wdict[wflow] = rdict
        return wdict

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Perform the following logic:
        - obtain list of current active workflows from the agent
        - obtain list of workflows not yet active but already in local workqueue
        - requests their specs from upstream wmstats server
        - update site lists of active workflows
        - push new specs to the agent local WorkQueue and update pickle spec file

        :return: none
        """
        # get list of active workflows from the agent, the returned dict
        # is composed by workflow names and associated pickle file (data comes from WMBS)
        wflows = self.getActiveWorkflows()
        self.logger.info("This agent has %d active workflows", len(wflows))
        self.logger.debug("Active workflows in the agent are: %s", wflows)

        # find workflows not yet active but already in local workqueue
        # note that some workflows here might already be active as well (multi WQEs)
        unacquiredWflows = self.findUnacquiredWorkflows()
        self.logger.info("This agent has %d not yet active workflows in local workqueue", len(unacquiredWflows))

        # now join all active and potentially not-yet-active workflows
        wflows = list(set(wflows + unacquiredWflows))

        # obtain workflow records from wmstats server
        wdict = self.wmstatsDict(wflows)
        self.logger.info("There is a total of %d common active workflows in the agent and wmstats", len(wdict))

        # iterate over the list of active workflows which is smaller than list from wmstats
        for wflow in wflows:
            if wflow not in wdict.keys():
                continue
            siteWhiteList = wdict[wflow]['SiteWhitelist']
            siteBlackList = wdict[wflow]['SiteBlacklist']

            # get the local Workqueue url for the workflow's spec
            specUrl = self.localWQ.hostWithAuth + "/%s/%s/spec" % (self.localWQ.db.name, wflow)

            # create wrapper helper and load the spec from local couch
            wHelper = WMWorkloadHelper()
            wHelper.load(specUrl)

            # extract from pickle spec both white and black site lists and compare them
            # to one we received from upstream service (ReqMgr2)
            wmaWhiteList = wHelper.getSiteWhitelist()
            wmaBlackList = wHelper.getSiteBlacklist()
            if set(wmaWhiteList) != set(siteWhiteList) or set(wmaBlackList) != set(siteBlackList):
                self.logger.info("Updating %s:", wflow)
                self.logger.info("  siteWhitelist %s => %s", wmaWhiteList, siteWhiteList)
                self.logger.info("  siteBlacklist %s => %s", wmaBlackList, siteBlackList)
                try:
                    # update local WorkQueue first
                    params = {'SiteWhitelist': siteWhiteList, 'SiteBlacklist': siteBlackList}
                    self.localWQ.updateElementsByWorkflow(wHelper, params, status=self.wqeStates)
                    msg = f"Successfully updated elements for workflow '{wflow}', "
                    msg += f"under WQ states: {self.wqeStates} and spec at: {sanitizeURL(specUrl).get('url')}"
                    self.logger.info(msg)
                except Exception as ex:
                    logging.exception("Unexpected exception while updating elements in local workqueue Details:\n%s", str(ex))
                    continue
            else:
                self.logger.info("No site list changes found for workflow %s", wflow)



"""
ReqMgr request handling.

"""
from __future__ import print_function
from future.utils import viewitems, viewvalues

from hashlib import md5
from copy import deepcopy

from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytesConditional
from WMCore.Lexicon import procdataset
from WMCore.REST.Auth import authz_match
from WMCore.ReqMgr.DataStructs.Request import initialize_request_args, initialize_clone
from WMCore.ReqMgr.DataStructs.RequestError import InvalidStateTransition, InvalidSpecParameterValue
from WMCore.ReqMgr.DataStructs.RequestStatus import (check_allowed_transition, get_modifiable_properties,
                                STATES_ALLOW_ONLY_STATE_TRANSITION, ALLOWED_STAT_KEYS)
from WMCore.ReqMgr.Tools.cms import releases, architectures, dashboardActivities
from WMCore.Services.DBS.DBS3Reader import getDataTiers
from WMCore.WMFactory import WMFactory
from WMCore.WMSpec.StdSpecs.StdBase import StdBase
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper
from WMCore.WMSpec.WMWorkloadTools import loadSpecClassByType, setArgumentsWithDefault
from WMCore.Cache.GenericDataCache import GenericDataCache, MemoryCacheStruct


def workqueue_stat_validation(request_args):
    return set(request_args.keys()) == set(ALLOWED_STAT_KEYS)


def _validate_request_allowed_args(reqArgs, newReqArgs):
    """
    Compares two request configuration dictionaries and returns the difference
    between them, but only at the first level (no recursive operations are attempted).
    Returns the key/value pairs taken from the newReqArgs only if they are allowed
    for the given request status.
    :param reqArgs:    A dictionary with the current request definition
    :param newReqArgs: A dictionary with user-provided request parameter changes
    :return:           A dictionary reflecting the difference between the above two
    NOTE:    This is asymmetrical operation/comparison, where newReqArs' values
             are considered to take precedence but not the other way around. They
             are in fact checked if they differ from the values at reqArgs and only those
             items which differ are returned.
    """
    reqArgsDiff = {}
    status = reqArgs["RequestStatus"]

    # Checking all keys that differ:
    reqArgsDiffKeys = []
    for key in newReqArgs:
        if key in reqArgs and newReqArgs[key] == reqArgs[key]:
            continue
        else:
            reqArgsDiffKeys.append(key)

    allowedKeys = deepcopy(get_modifiable_properties(status))
    allowedKeys.extend(ALLOWED_STAT_KEYS)

    # Filter out all fields which are not allowed for the given source status:
    for key in reqArgsDiffKeys:
        if key in allowedKeys:
            reqArgsDiff[key] = newReqArgs[key]

    return reqArgsDiff


def validate_request_update_args(request_args, config, reqmgr_db_service, param):
    """
    param and safe structure is RESTArgs structure: named tuple
    RESTArgs(args=[], kwargs={})

    validate post/put request
    1. read data from body
    2. validate the permission (authentication)
    3. validate state transition (against previous state from couchdb)
    2. validate using workload validation
    3. convert data from body to arguments (spec instance, argument with default setting)

    TODO: raise right kind of error with clear message
    """
    # this needs to be deleted for validation
    request_name = request_args.pop("RequestName")

    couchurl = '%s/%s' % (config.couch_host, config.couch_reqmgr_db)
    workload = WMWorkloadHelper()
    workload.loadSpecFromCouch(couchurl, request_name)

    # validate the status
    # NOTE: For state transition validation we do not consider request parameter
    #       difference, but we rather act against the request parameter changes
    #       as they have been requested
    if "RequestStatus" in request_args:
        validate_state_transition(reqmgr_db_service, request_name, request_args["RequestStatus"])
        if request_args["RequestStatus"] in STATES_ALLOW_ONLY_STATE_TRANSITION:
            # if state change doesn't allow other transition nothing else to validate
            args_only_status = {}
            args_only_status["RequestStatus"] = request_args["RequestStatus"]
            args_only_status["cascade"] = request_args.get("cascade", False)
            return workload, args_only_status
        elif request_args["RequestStatus"] == 'assigned':
            workload.validateArgumentForAssignment(request_args)
        validate_request_priority(request_args)
        return workload, request_args
    else:
        request = reqmgr_db_service.getRequestByNames(request_name)
        request = request[request_name]
        workload.setStatus(request['RequestStatus'])
        # NOTE: Even though later we return only the arguments which have changed during the call,
        #       we first need to validate the full set of arguments as provided  by the user, because
        #       the combination matters. e.g. Adding a site to SiteBlacklist, while keeping it in the
        #       SiteWhitelist as well. In such a case only the SiteBlacklist parameter will enter the
        #       list of changed  arguments - reqargsDiff, hence the validation would miss the collision
        # NOTE: We also must skip validating update stat arguments
        if not workqueue_stat_validation(request_args):
            workload.validateArgumentsPartialUpdate(request_args)
        reqArgsDiff = _validate_request_allowed_args(request, request_args)
        validate_request_priority(reqArgsDiff)
        return workload, reqArgsDiff


def validate_request_priority(reqArgs):
    """
    Validate the RequestPriority argument against its definition
    in StdBase
    :param reqArgs: dictionary of user request arguments
    :return: nothing, but raises an exception in case of an invalid value
    """
    if 'RequestPriority' in reqArgs:
        reqPrioDefin = StdBase.getWorkloadCreateArgs()['RequestPriority']
        if not isinstance(reqArgs['RequestPriority'], reqPrioDefin['type']):
            msg = "RequestPriority must be of integer type, not: {}".format(type(reqArgs['RequestPriority']))
            raise InvalidSpecParameterValue(msg)
        if reqPrioDefin['validate'](reqArgs['RequestPriority']) is False:
            raise InvalidSpecParameterValue("RequestPriority must be an integer between 0 and 999999")


def validate_request_create_args(request_args, config, reqmgr_db_service, *args, **kwargs):
    """
    *arg and **kwargs are only for the interface
    validate post request
    1. read data from body
    2. validate using spec validation
    3. convert data from body to arguments (spec instance, argument with default setting)
    TODO: raise right kind of error with clear message
    """
    if request_args["RequestType"] == "Resubmission":
        # do not set default values for Resubmission since it will be inherited from parent
        # both create & assign args are accepted for Resubmission creation
        workload, request_args = validate_resubmission_create_args(request_args, config, reqmgr_db_service)
    else:
        initialize_request_args(request_args, config)

        # load the correct class in order to validate the arguments
        specClass = loadSpecClassByType(request_args["RequestType"])
        # set default values for the request_args
        setArgumentsWithDefault(request_args, specClass.getWorkloadCreateArgs())
        spec = specClass()
        workload = spec.factoryWorkloadConstruction(request_args["RequestName"],
                                                    request_args)

    return workload, request_args


def validate_resubmission_create_args(request_args, config, reqmgr_db_service, *args, **kwargs):
    """
    Handle resubmission workflows, loading the spec arguments definition (and chain
    definition, if needed) and inheriting all arguments defined in the spec.
    User can also override arguments, since it uses the same mechanism as a clone.
    *arg and **kwargs are only for the interface
    """
    response = reqmgr_db_service.getRequestByNames(request_args["OriginalRequestName"])
    originalArgs = next(iter(viewvalues(response)))

    ### not a nice fix for #8245, but we cannot inherit the CollectionName attr
    originalArgs.pop("CollectionName", None)

    chainArgs = None
    if originalArgs["RequestType"] == 'Resubmission':
        # ACDC of ACDC, we can't validate this case
        # simply copy the whole original dictionary over and accept all args
        createArgs = originalArgs
        request_args["OriginalRequestType"] = originalArgs["OriginalRequestType"]
        request_args["ResubmissionCount"] = originalArgs.get("ResubmissionCount", 1) + 1
    else:
        # load arguments definition from the proper/original spec factory
        parentClass = loadSpecClassByType(originalArgs["RequestType"])
        createArgs = parentClass.getWorkloadAssignArgs()
        if originalArgs["RequestType"] in ('StepChain', 'TaskChain'):
            chainArgs = parentClass.getChainCreateArgs()
        createArgs.update(parentClass.getWorkloadCreateArgs())
        request_args["OriginalRequestType"] = originalArgs["RequestType"]

    cloned_args = initialize_clone(request_args, originalArgs, createArgs, chainArgs)
    initialize_request_args(cloned_args, config)

    specClass = loadSpecClassByType(request_args["RequestType"])
    spec = specClass()
    workload = spec.factoryWorkloadConstruction(cloned_args["RequestName"],
                                                cloned_args, request_args)

    return workload, cloned_args


def validate_clone_create_args(request_args, config, reqmgr_db_service, *args, **kwargs):
    """
    Handle clone workflows through the clone API, by loading the spec arguments
    definition (and chain definition, if needed) and inheriting all arguments defined
    in the spec.
    *arg and **kwargs are only for the interface
    """
    response = reqmgr_db_service.getRequestByNames(request_args.pop("OriginalRequestName"))
    originalArgs = next(iter(viewvalues(response)))

    chainArgs = None
    specClass = loadSpecClassByType(originalArgs["RequestType"])
    if originalArgs["RequestType"] == 'Resubmission':
        # cloning an ACDC, nothing that we can validate
        # simply copy the whole original dictionary over and accept all args
        createArgs = originalArgs
    else:
        # load arguments definition from the proper/original spec factory
        createArgs = specClass.getWorkloadCreateArgs()
        if originalArgs["RequestType"] in ('StepChain', 'TaskChain'):
            chainArgs = specClass.getChainCreateArgs()

    cloned_args = initialize_clone(request_args, originalArgs, createArgs, chainArgs)
    initialize_request_args(cloned_args, config)

    spec = specClass()
    workload = spec.factoryWorkloadConstruction(cloned_args["RequestName"], cloned_args)

    return workload, cloned_args


def validate_state_transition(reqmgr_db_service, request_name, new_state):
    """
    validate state transition by getting the current data from
    couchdb
    """
    requests = reqmgr_db_service.getRequestByNames(request_name)
    # generator object can't be subscribed: need to loop.
    # only one row should be returned
    for request in viewvalues(requests):
        current_state = request["RequestStatus"]
    if not check_allowed_transition(current_state, new_state):
        raise InvalidStateTransition(request_name, current_state, new_state)
    return


def create_json_template_spec(specArgs):
    template = {}
    for key, prop in viewitems(specArgs):

        if key == "RequestorDN":
            # this will be automatically collected so skip it.
            continue

        if key == "CMSSWVersion":
            # get if from tag collector
            value = releases()
        elif key == "ScramArch":
            value = architectures()
        elif prop == "Dashboard":
            value = dashboardActivities()
        elif prop.get("optional", True):
            # if optional need to always have default value
            value = prop["default"]
        else:
            value = "REPLACE-%s" % key
        template[key] = value
    return template


def get_request_template_from_type(request_type, loc="WMSpec.StdSpecs"):
    pluginFactory = WMFactory("specArgs", loc)
    alteredClassName = "%sWorkloadFactory" % request_type
    spec = pluginFactory.loadObject(classname=request_type, alteredClassName=alteredClassName)
    specArgs = spec.getWorkloadCreateArgs()

    result = create_json_template_spec(specArgs)
    return result


def validateOutputDatasets(outDsets, dbsUrl):
    """
    Validate output datasets after all the other arguments have been
    locally update during assignment.
    """
    if len(outDsets) != len(set(outDsets)):
        msg = "Output dataset contains duplicates and it has to be fixed! %s" % outDsets
        raise InvalidSpecParameterValue(msg)

    datatier = []
    for dataset in outDsets:
        procds, tier = dataset.split("/")[2:]
        datatier.append(tier)
        try:
            procdataset(procds)
        except AssertionError as ex:
            msg = "Bad output dataset name, check the processed dataset name.\n %s" % str(ex)
            raise InvalidSpecParameterValue(msg) from ex

    # TODO: this url conversion below can be removed in one year from now, thus March 2022
    dbsUrl = dbsUrl.replace("cmsweb.cern.ch", "cmsweb-prod.cern.ch")

    # Verify whether the output datatiers are available in DBS
    _validateDatatier(datatier, dbsUrl)


def _validateDatatier(datatier, dbsUrl, expiration=3600):
    """
    _validateDatatier_

    Provided a list of datatiers extracted from the outputDatasets, checks
    whether they all exist in DBS.
    """
    cacheName = "dataTierList_" + md5(encodeUnicodeToBytesConditional(dbsUrl, condition=PY3)).hexdigest()
    if not GenericDataCache.cacheExists(cacheName):
        mc = MemoryCacheStruct(expiration, getDataTiers, kwargs={'dbsUrl': dbsUrl})
        GenericDataCache.registerCache(cacheName, mc)

    cacheData = GenericDataCache.getCacheData(cacheName)
    dbsTiers = cacheData.getData()
    badTiers = list(set(datatier) - set(dbsTiers))
    if badTiers:
        raise InvalidSpecParameterValue("Bad datatier(s): %s not available in DBS." % badTiers)


def isUserAllowed(authzCls, requestArgs):
    """
    Checks whether user is allowed to perform a given action,
    based on the request arguments provided and the user's
    roles and groups
    :param authzCls: an instance of the AuthzByStatus class
    :param requestArgs: dictionary with the arguments provided by the user
    :return: raise an HTTPError if the user is not allowed to
        perform action, otherwise return None
    """
    permission = authzCls.getRolesGroupsByStatus(requestArgs)
    authz_match(permission['role'], permission['group'])



#!/usr/bin/env python
"""
_JobFactory_

"""

from builtins import range

import logging
import threading
import operator

from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.WMObject import WMObject
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File as WMBSFile
from WMCore.WMExceptions import WM_JOB_ERROR_CODES


class JobFactory(WMObject):
    """
    A JobFactory is created with a subscription (which has a fileset). It is a
    base class of an object instance of an object representing some job
    splitting algorithm. It is called with a job type (at least) to return a
    JobGroup object. The JobFactory should be subclassed by real splitting
    algorithm implementations.
    """

    def __init__(self, package='WMCore.DataStructs', subscription=None, generators=None, limit=0):
        super(JobFactory, self).__init__()
        self.package = package
        self.subscription = subscription
        self.generators = generators if generators else []
        self.jobInstance = None
        self.groupInstance = None
        self.jobGroups = []
        self.currentGroup = None
        self.currentJob = None
        self.nJobs = 0
        self.baseUUID = None
        self.limit = limit
        self.transaction = None
        self.proxies = []
        self.grabByProxy = False
        self.daoFactory = None
        self.timing = {'jobInstance': 0, 'sortByLocation': 0, 'acquireFiles': 0, 'jobGroup': 0}
        self.siteWhitelist = []
        self.siteBlacklist = []
        self.trustSitelists = False
        self.trustPUSitelists = False

        if package == "WMCore.WMBS":
            myThread = threading.currentThread()

            self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                         logger=myThread.logger,
                                         dbinterface=myThread.dbi)
            self.getParentInfoAction = self.daoFactory(classname="Files.GetParentAndGrandParentInfo")

            self.pnn_to_psn = self.daoFactory(classname="Locations.GetPNNtoPSNMapping").execute()

        return

    def __call__(self, jobtype="Job", grouptype="JobGroup", *args, **kwargs):
        """
        __call__


        """

        # Need to reset the internal data for multiple calls to the factory
        self.jobGroups = []
        self.currentGroup = None
        self.currentJob = None

        self.siteWhitelist = kwargs.get("siteWhitelist", [])
        self.siteBlacklist = kwargs.get("siteBlacklist", [])
        self.trustSitelists = kwargs.get("trustSitelists", False)
        self.trustPUSitelists = kwargs.get("trustPUSitelists", False)

        # Every time we restart, re-zero the jobs
        self.nJobs = 0

        # Create a new name
        self.baseUUID = makeUUID()

        module = "%s.%s" % (self.package, jobtype)
        module = __import__(module, globals(), locals(), [jobtype])  # , -1)
        self.jobInstance = getattr(module, jobtype.split('.')[-1])

        module = "%s.%s" % (self.package, grouptype)
        module = __import__(module, globals(), locals(), [grouptype])
        self.groupInstance = getattr(module, grouptype.split('.')[-1])

        list([x.start() for x in self.generators])

        self.limit = int(kwargs.get("file_load_limit", self.limit))
        self.algorithm(*args, **kwargs)
        self.commit()

        list([x.finish() for x in self.generators])
        return self.jobGroups

    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        A splitting algorithm that takes all available files from the
        subscription and splits them into jobs and inserts them into job groups.
        The algorithm must return a list of job groups.
        """
        self.newGroup()
        self.newJob(name='myJob')
        return

    def newGroup(self):
        """
        Return and new JobGroup
        """
        self.appendJobGroup()
        self.currentGroup = self.groupInstance(subscription=self.subscription)
        list([x.startGroup(self.currentGroup) for x in self.generators])
        return

    def newJob(self, name=None, files=None, failedJob=False, failedReason=None, failedErrCode=None):
        """
        Instantiate a new Job onject, apply all the generators to it
        """
        if name is None:
            name = self.getJobName()
        self.currentJob = self.jobInstance(name, files)
        self.currentJob["task"] = self.subscription.taskName()
        self.currentJob["workflow"] = self.subscription.workflowName()
        self.currentJob["jobType"] = self.subscription["type"]
        self.currentJob["taskType"] = self.subscription.workflowType()
        self.currentJob["owner"] = self.subscription.owner()

        # All production jobs must be run 1
        if self.subscription["type"] == "Production":
            self.currentJob["mask"].setMaxAndSkipRuns(0, 1)

        # Some jobs are not meant to be submitted, ever
        if failedJob:
            self.currentJob["failedOnCreation"] = True
            self.currentJob["failedReason"] = failedReason
            self.currentJob["failedErrCode"] = failedErrCode

        # Decides how the pileup data will be handled in runtime
        if self.trustPUSitelists:
            self.currentJob.addBaggageParameter("trustPUSitelists", self.trustPUSitelists)

        self.nJobs += 1
        for gen in self.generators:
            gen(self.currentJob)
        self.currentGroup.add(self.currentJob)
        return

    def appendJobGroup(self):
        """
        Append jobGroup to jobGroup list

        """

        if self.currentGroup:
            list([x.finishGroup(self.currentGroup) for x in self.generators])
        if self.currentGroup:
            self.jobGroups.append(self.currentGroup)
            self.currentGroup = None

        return

    def commit(self):
        """
        Bulk commit the JobGroups all at once
        """
        self.appendJobGroup()

        if len(self.jobGroups) == 0:
            return

        logging.debug("About to commit %i jobGroups", len(self.jobGroups))
        logging.debug("About to commit %i jobs", len(self.jobGroups[0].newjobs))

        if self.package == 'WMCore.WMBS':

            for jobGroup in self.jobGroups:

                for job in jobGroup.newjobs:
                    # temporary place holder for file location
                    fileLocations = set([])
                    if self.trustSitelists:
                        locSet = set(self.siteWhitelist) - set(self.siteBlacklist)
                    else:
                        locSet = set([])
                        for pnn in job['input_files'][0]['locations']:
                            locSet.update(self.pnn_to_psn.get(pnn, []))
                        fileLocations = locSet
                        if len(self.siteWhitelist) > 0:
                            locSet = locSet & set(self.siteWhitelist)
                        if len(self.siteBlacklist) > 0:
                            locSet = locSet - set(self.siteBlacklist)

                    job['possiblePSN'] = locSet
                    if len(job['possiblePSN']) == 0:
                        # NOTE: If we have no place to execute a single job we mark it as failedOnCreation.
                        #       We have two options:
                        #       * Either propagate the job state to 'createdfailed' here immediately
                        #       * Or we let JobCreatorPoller later catch the marker and do it for us
                        #         when moving the jobs from state 'new' to 'created'
                        #       For the time being we opt for the later
                        job["failedOnCreation"] = True
                        job["failedErrCode"] = 71101
                        job["failedReason"] = WM_JOB_ERROR_CODES[job["failedErrCode"]]
                        msg = "Found job with an empty list of possiblePSN. JobId: %s from JobGroup: %s and task: %s."
                        msg += WM_JOB_ERROR_CODES[job["failedErrCode"]]
                        msg += " Marking it as FailedOnCreation"
                        logging.warning(msg, job['id'], job['jobgroup'], job['task'] )

                        job["fileLocations"] = fileLocations
                        job["siteWhitelist"] = self.siteWhitelist
                        job["siteBlacklist"] = self.siteBlacklist

                # now after the jobs are created, remove input file locations
                # they are no longer needed and just take up space
                for job in jobGroup.newjobs:
                    for fileInfo in job['input_files']:
                        fileInfo['locations'] = set([])

            self.subscription.bulkCommit(jobGroups=self.jobGroups)

        else:

            # we have a DataStructs job and have to do everything by hand
            for jobGroup in self.jobGroups:
                jobGroup.commit()
                for job in jobGroup.jobs:
                    job.save()
            self.subscription.save()

        return

    def sortByLocation(self):
        """
        _sortByLocation_

        Retrieve available files and return them sorted by location.
        The keys in the dict correspond to a set of locations.

        If TrustSiteLists is set ignore the file locations and treat all files
        as being present in the same place (use key "AAA" location).
        """

        fileDict = {}

        if self.grabByProxy:
            logging.debug("About to load files by proxy")
            fileset = self.loadFiles(size=self.limit)
            logging.debug("Loaded %i files", len(fileset))
        else:
            logging.debug("About to load files by DAO")
            fileset = self.subscription.availableFiles(limit=self.limit, doingJobSplitting=True)

        # Reverse sorting this set by location is required to match how sorting was done in py2
        # Unittests that rely on this sorting method
        # - WMCore_t.WMBS_t.JobSplitting_t.FileBased_t.FileBasedTest:testSiteWhiteBlacklist
        # - WMCore_t.WMBS_t.JobSplitting_t.FileBased_t.FileBasedTest:testLocationSplit
        for fileInfo in sorted(fileset, key=operator.itemgetter('locations'), reverse=True):

            if self.trustSitelists:
                locSet = frozenset(set(['AAA']))
            else:
                locSet = frozenset(fileInfo['locations'])

                if len(locSet) == 0:
                    logging.error("File %s has no locations!", fileInfo['lfn'])

            if locSet in fileDict:
                fileDict[locSet].append(fileInfo)
            else:
                fileDict[locSet] = [fileInfo]

        return fileDict

    def getJobName(self, length=None):
        """
        _getJobName_

        Creates a job name based on workflow and task
        Uses a passed in integer length that MUST be unique!
        """

        if not length:
            length = self.nJobs

        name = '%s-%i' % (self.baseUUID, length)

        return name

    def open(self):
        """
        _open_

        Open a connection to the database, and put
        resulting ResultProxies in self.proxies
        """

        logging.debug("Opening DB resultProxies for JobFactory")

        myThread = threading.currentThread()

        subAction = self.daoFactory(classname="Subscriptions.GetAvailableFilesNoLocations")
        results = subAction.execute(subscription=self.subscription['id'],
                                    returnCursor=True,
                                    conn=myThread.transaction.conn,
                                    transaction=True)

        for proxy in results:
            self.proxies.append(proxy)

        logging.debug("Received %i proxies", len(self.proxies))
        # Activate everything so that we grab files by proxy
        self.grabByProxy = True

        return

    def close(self):
        """
        _close_

        Close any leftover connections
        """
        self.proxies = []
        self.grabByProxy = False
        return

    def loadFiles(self, size=10):
        """
        _loadFiles_

        Grab some files from the resultProxy
        Should handle multiple proxies.  Not really sure about that
        """

        if len(self.proxies) < 1:
            # Well, you don't have any proxies.
            # This is what happens when you ran out of files last time
            logging.info("No additional files found; Ending.")
            return set()

        resultProxy = self.proxies[0]
        rawResults = []
        if isinstance(resultProxy.keys, list):
            keys = resultProxy.keys
        else:
            # the object below is of the type: sqlalchemy.engine.result.RMKeyView
            keys = list(resultProxy.keys())
        files = set()

        while len(rawResults) < size and len(self.proxies) > 0:
            length = size - len(rawResults)
            newResults = resultProxy.fetchmany(size=length)
            if len(newResults) < length:
                # Assume we're all out
                # Eliminate this proxy
                self.proxies.remove(resultProxy)
            rawResults.extend(newResults)

        if rawResults == []:
            # Nothing to do
            return set()

        fileList = self.formatDict(results=rawResults, keys=keys)
        fileIDs = list(set([x['fileid'] for x in fileList]))

        myThread = threading.currentThread()
        fileInfoAct = self.daoFactory(classname="Files.GetForJobSplittingByID")
        fileInfoDict = fileInfoAct.execute(file=fileIDs,
                                           conn=myThread.transaction.conn,
                                           transaction=True)

        getLocAction = self.daoFactory(classname="Files.GetLocationBulk")
        getLocDict = getLocAction.execute(files=fileIDs,
                                          conn=myThread.transaction.conn,
                                          transaction=True)

        for fID in fileIDs:
            fl = WMBSFile(id=fID)
            fl.update(fileInfoDict[fID])
            locations = getLocDict.get((fID), [])
            for loc in locations:
                fl.setLocation(loc, immediateSave=False)
            files.add(fl)

        return files

    def formatDict(self, results, keys):
        """
        _formatDict_

        Cast the file column to an integer as the DBFormatter's formatDict()
        method turns everything into strings.  Also, fixup the results of the
        Oracle query by renaming 'fileid' to file.
        """

        formattedResults = []
        for entry in results:
            indivDict = {}
            for i in range(len(keys)):
                # Assign each key to a value
                indivDict[str(keys[i]).lower()] = entry[i]
            formattedResults.append(indivDict)

        return formattedResults

    def findParent(self, lfn):
        """
        _findParent_

        Find the parents for a file based on its lfn
        """

        parentsInfo = self.getParentInfoAction.execute([lfn])
        newParents = set()
        for parentInfo in parentsInfo:

            # This will catch straight to merge files that do not have redneck
            # parents.  We will mark the straight to merge file from the job
            # as a child of the merged parent.
            if int(parentInfo["merged"]) == 1:
                newParents.add(parentInfo["lfn"])

            elif parentInfo['gpmerged'] is None:
                continue

            # Handle the files that result from merge jobs that aren't redneck
            # children.  We have to setup parentage and then check on whether or
            # not this file has any redneck children and update their parentage
            # information.
            elif int(parentInfo["gpmerged"]) == 1:
                newParents.add(parentInfo["gplfn"])

            # If that didn't work, we've reached the great-grandparents
            # And we have to work via recursion
            else:
                parentSet = self.findParent(lfn=parentInfo['gplfn'])
                for parent in parentSet:
                    newParents.add(parent)

        return newParents

    def checkForAmountOfWork(self):
        """
        Depending on what and how we are running the job splitting, check or
        not for a minimum amount of work (lumi sections) before passing the list
        of files all the way to the algorithm
        :return: a boolean flag
        """
        if self.collectionName:
            return False

        fileset = self.subscription.getFileset()
        fileset.load()
        if not fileset.open:
            return False  # it's closed, so just run the last jobs for it

        return True

    def getFilesSortedByLocation(self, eventsPerJob):
        """
        _getFilesSortedByLocation_

        Function used by the EventAware* splitting algorithms for retrieving
        a list of files available and sort them by location.

        If the fileset is closed, keep splitting these data. Otherwise check
        whether there are enough events in each of these locations, if events
        don't match the desired events_per_job splitting parameter, then skip
        those files until further cycles.
        :param eventsPerJob: number of events desired in the splitting
        :return: a dictionary of files, key'ed by a frozenset location
        """
        lDict = self.sortByLocation()
        if not self.loadRunLumi:
            return lDict  # then it's a DataStruct/CRAB splitting

        if self.checkForAmountOfWork():
            # first, check whether we have enough files to reach the desired events_per_job
            for sites in list(lDict):  # lDict changes size during for loop!
                availableEventsPerLocation = sum([f['events'] for f in lDict[sites]])
                if eventsPerJob > availableEventsPerLocation:
                    # then we don't split these files for the moment
                    lDict.pop(sites)

        return lDict

    @staticmethod
    def getPerformanceParameters(defaultParams):
        """
        _getPerformanceParameters_

        This generic JobFactory function allows all splitters to easily
        retrieve the parameters to specify the resource requirements for
        the created jobs.
        # TODO: Check couchDB for performance on the current task
                and if not available it falls back to the default params.
        It returns a tuple with the following values:
        timePerEvent, sizePerEvent, memoryRequirement
        """
        timePerEvent = defaultParams.get('timePerEvent', 0) or 0
        sizePerEvent = defaultParams.get('sizePerEvent', 0) or 0
        memory = defaultParams.get('memoryRequirement', 0) or 0
        return timePerEvent, sizePerEvent, memory



from __future__ import (division, print_function)
from builtins import object


class MockApMon(object):
    """
    Version of Services/PhEDEx intended to be used with mock or unittest.mock
    """
    def __init__(self, *args, **kwargs):
        print("Using MockApMon")

    def sendParams(self, params):
        """
        Send multiple parameters to MonALISA, with default(last given) cluser and node names.
        """
        return

    def sendTimedParams(self, timeStamp, params):
        """
        Send multiple parameters, specifying the time for them, with default(last given) cluster and node names.
        (See sendTimedParameters for more details)
        """
        return

    def sendParameter(self, clusterName, nodeName, paramName, paramValue):
        """
        Send a single parameter to MonALISA.
        """
        return

    def sendTimedParameter(self, clusterName, nodeName, timeStamp, paramName, paramValue):
        """
        Send a single parameter, with a given time.
        """
        return

    def sendParameters(self, clusterName, nodeName, params):
        """
        Send multiple parameters specifying cluster and node name for them
        """
        return

    def sendTimedParameters(self, clusterName, nodeName, timeStamp, params):
        """
        Send multiple monitored parameters to MonALISA.

        - clusterName is the name of the cluster being monitored. The first
          time this function is called, this paramenter must not be None. Then,
          it can be None; last given clusterName will be used instead.
        - nodeName is the name of the node for which are the parameters. If this
          is None, the full hostname of this machine will be sent instead.
        - timeStamp, if > 0, is given time for the parameters. This is in seconds from Epoch.
          Note that this option should be used only if you are sure about the time for the result.
          Otherwize, the parameters will be assigned a correct time(obtained from NTP servers)
          in MonALISA service. This option can be usefull when parsing logs, for example.
        - params is a dictionary containing pairs with:
            - key: parameter name
            - value: parameter value, either int or float.
          or params is a vector of tuples(key, value). This version can be used
          in case you want to send the parameters in a given order.

        NOTE that python doesn't know about 32-bit floats(only 64-bit floats!)
        """
        return

    def addJobToMonitor(self, pid, workDir, clusterName, nodeName):
        """
        Add a new job to monitor.
        """
        return

    def removeJobToMonitor(self, pid):
        """
        Remove a job from being monitored.
        """
        return

    def setMonitorClusterNode(self, clusterName, nodeName):
        """
        Set the cluster and node names where to send system related information.
        """
        return

    def enableBgMonitoring(self, onOff):
        """
        Enable or disable background monitoring. Note that background monitoring information
        can still be sent if user calls the sendBgMonitoring method.
        """
        return

    def sendBgMonitoring(self, mustSend=False):
        """
        Send background monitoring about system and jobs to all interested destinations.
        If mustSend == True, the information is sent regardles of the elapsed time since last sent
        If mustSend == False, the data is sent only if the required interval has passed since last sent
        """
        return

    def setDestinations(self, initValue):
        """
        Set the destinations of the ApMon instance. It accepts the same parameters as the constructor.
        """
        return

    def getConfig(self):
        """
        Returns a multi-line string that contains the configuration of ApMon. This string can
        be passed to the setDestination method(or to the constructor). It has the same
        structure as the config file/url contents.
        """

        return {}

    def initializedOK(self):
        """
        Returns true if there are destination(s) configured.
        """
        return True

    def freedOK(self):
        """
        Returns true if all ApMon resources were properly freed.
        """
        return True

    def setLogLevel(self, strLevel):
        """
        Change the log level. Given level is a string, one of 'FATAL', 'ERROR', 'WARNING',
        'INFO', 'NOTICE', 'DEBUG'.
        """
        return

    def setMaxMsgRate(self, rate):
        """
        Set the maximum number of messages that can be sent, per second.
        """
        return

    def setMaxMsgSize(self, size):
        """
        Set the maximum size of the sent messages. ApMon will try to split in several independent
        messages parameters sent in bulk, if the size would be larger than this
        """
        return

    def free(self):
        """
        Stop background threands, close opened sockets. You have to use this function if you want to
        free all the resources that ApMon takes, and allow it to be garbage-collected.
        """
        return



#!/usr/bin/env python
"""
_BulkAdd_

Oracle implementation of Fileset.BulkAdd
"""

from WMCore.WMBS.MySQL.Fileset.BulkAdd import BulkAdd as BulkAddFilesetMySQL

class BulkAdd(BulkAddFilesetMySQL):
    pass



#!/usr/bin/env python
"""
_SetBlocksRule_

MySQL implementation of RucioInjector.Database.SetBlocksRule
"""

from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter


class SetBlocksRule(DBFormatter):
    """
    _SetBlocksRule_

    Set rules ID for blocks in the DBSBuffer
    """
    sql = """UPDATE dbsbuffer_block
               SET dbsbuffer_block.rule_id = :RULE_ID
             WHERE dbsbuffer_block.blockname = :BLOCKNAME
             """

    def execute(self, binds, conn=None, transaction=False):
        self.dbi.processData(self.sql, binds, conn=conn,
                             transaction=transaction)

        return



#!/usr/bin/env python
"""
_Repack_t_

Unit tests for the Tier0 Repack workflow.
"""

from __future__ import division, print_function
from future.utils import viewitems

import threading
import unittest
from copy import deepcopy

from Utils.PythonVersion import PY3

from WMCore.DAOFactory import DAOFactory
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMSpec.StdSpecs.Repack import RepackWorkloadFactory
from WMCore.WorkQueue.WMBSHelper import WMBSHelper
from WMQuality.TestInitCouchApp import TestInitCouchApp

REQUEST = {
    'MaxSizeSingleLumi': 12 * 1024 * 1024 * 1024,
    'MaxSizeMultiLumi': 8 * 1024 * 1024 * 1024,
    'MinInputSize': 2.1 * 1024 * 1024 * 1024,
    'MaxInputSize': 4 * 1024 * 1024 * 1024,
    'MaxEdmSize': 12 * 1024 * 1024 * 1024,
    'MaxOverSize': 8 * 1024 * 1024 * 1024,
    'MaxInputEvents': 3 * 1000 * 1000,
    'MaxInputFiles': 1000,
    'MaxLatency': 24 * 3600,
    'MinMergeSize': 2.1 * 1024 * 1024 * 1024,
    'MaxMergeEvents': 3 * 1000 * 1000,
    'RunNumber': 123456,
    'AcquisitionEra': "TestAcquisitionEra",
    'ValidStatus': "VALID",
    'Outputs': [{'dataTier': "RAW",
                 'eventContent': "All",
                 'selectEvents': ["Path1:HLT,Path2:HLT"],
                 'primaryDataset': "PrimaryDataset1"},
                {'dataTier': "RAW",
                 'eventContent': "All",
                 'selectEvents': ["Path3:HLT,Path4:HLT"],
                 'primaryDataset': "PrimaryDataset2"}]
}


class RepackTests(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Initialize the database and couch.
        """
        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)
        self.testDir = self.testInit.generateWorkDir()

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.listTasksByWorkflow = self.daoFactory(classname="Workflow.LoadFromName")
        self.listFilesets = self.daoFactory(classname="Fileset.List")
        self.listSubsMapping = self.daoFactory(classname="Subscriptions.ListSubsAndFilesetsFromWorkflow")
        if PY3:
            self.assertItemsEqual = self.assertCountEqual
        return

    def tearDown(self):
        """
        _tearDown_

        Clear out the database.
        """
        self.testInit.clearDatabase()
        self.testInit.delWorkDir()
        return

    def testRepack(self):
        """
        _testRepack_

        Create a Repack workflow
        and verify it installs into WMBS correctly.
        """
        testArguments = RepackWorkloadFactory.getTestArguments()
        testArguments.update(deepcopy(REQUEST))

        factory = RepackWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", testArguments)
        testWorkload.setSpecUrl("somespec")
        testWorkload.setOwnerDetails("Dirk.Hufnagel@cern.ch", "T0")

        testWMBSHelper = WMBSHelper(testWorkload, "Repack", cachepath=self.testDir)
        testWMBSHelper.createTopLevelFileset()
        testWMBSHelper._createSubscriptionsInWMBS(testWMBSHelper.topLevelTask, testWMBSHelper.topLevelFileset)

        repackWorkflow = Workflow(name="TestWorkload",
                                  task="/TestWorkload/Repack")
        repackWorkflow.load()
        self.assertEqual(len(repackWorkflow.outputMap), len(testArguments["Outputs"]) + 1,
                         "Error: Wrong number of WF outputs in the Repack WF.")

        goldenOutputMods = {"write_PrimaryDataset1_RAW": "RAW", "write_PrimaryDataset2_RAW": "RAW"}
        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            fset = goldenOutputMod + tier
            mergedOutput = repackWorkflow.outputMap[fset][0]["merged_output_fileset"]
            unmergedOutput = repackWorkflow.outputMap[fset][0]["output_fileset"]
            mergedOutput.loadData()
            unmergedOutput.loadData()

            if goldenOutputMod != "write_PrimaryDataset1_RAW":
                self.assertEqual(mergedOutput.name,
                                 "/TestWorkload/Repack/RepackMerge%s/merged-Merged%s" % (goldenOutputMod, tier),
                                 "Error: Merged output fileset is wrong: %s" % mergedOutput.name)
            self.assertEqual(unmergedOutput.name, "/TestWorkload/Repack/unmerged-%s" % fset,
                             "Error: Unmerged output fileset is wrong: %s" % unmergedOutput.name)

        logArchOutput = repackWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
        unmergedLogArchOutput = repackWorkflow.outputMap["logArchive"][0]["output_fileset"]
        logArchOutput.loadData()
        unmergedLogArchOutput.loadData()

        self.assertEqual(logArchOutput.name, "/TestWorkload/Repack/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")
        self.assertEqual(unmergedLogArchOutput.name, "/TestWorkload/Repack/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            mergeWorkflow = Workflow(name="TestWorkload",
                                     task="/TestWorkload/Repack/RepackMerge%s" % goldenOutputMod)
            mergeWorkflow.load()

            self.assertEqual(len(mergeWorkflow.outputMap), 3,
                             "Error: Wrong number of WF outputs.")

            mergedMergeOutput = mergeWorkflow.outputMap["Merged%s" % tier][0]["merged_output_fileset"]
            unmergedMergeOutput = mergeWorkflow.outputMap["Merged%s" % tier][0]["output_fileset"]

            mergedMergeOutput.loadData()
            unmergedMergeOutput.loadData()

            self.assertEqual(mergedMergeOutput.name,
                             "/TestWorkload/Repack/RepackMerge%s/merged-Merged%s" % (goldenOutputMod, tier),
                             "Error: Merged output fileset is wrong.")
            self.assertEqual(unmergedMergeOutput.name,
                             "/TestWorkload/Repack/RepackMerge%s/merged-Merged%s" % (goldenOutputMod, tier),
                             "Error: Unmerged output fileset is wrong.")

            logArchOutput = mergeWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
            unmergedLogArchOutput = mergeWorkflow.outputMap["logArchive"][0]["output_fileset"]
            logArchOutput.loadData()
            unmergedLogArchOutput.loadData()

            self.assertEqual(logArchOutput.name,
                             "/TestWorkload/Repack/RepackMerge%s/merged-logArchive" % goldenOutputMod,
                             "Error: LogArchive output fileset is wrong: %s" % logArchOutput.name)
            self.assertEqual(unmergedLogArchOutput.name,
                             "/TestWorkload/Repack/RepackMerge%s/merged-logArchive" % goldenOutputMod,
                             "Error: LogArchive output fileset is wrong.")

        topLevelFileset = Fileset(name="TestWorkload-Repack")
        topLevelFileset.loadData()

        repackSubscription = Subscription(fileset=topLevelFileset, workflow=repackWorkflow)
        repackSubscription.loadData()

        self.assertEqual(repackSubscription["type"], "Repack",
                         "Error: Wrong subscription type.")
        self.assertEqual(repackSubscription["split_algo"], "Repack",
                         "Error: Wrong split algorithm. %s" % repackSubscription["split_algo"])

        unmergedOutputs = {"write_PrimaryDataset1_RAW": "RAW", "write_PrimaryDataset2_RAW": "RAW"}
        for unmergedOutput, tier in viewitems(unmergedOutputs):
            fset = unmergedOutput + tier
            unmergedDataTier = Fileset(name="/TestWorkload/Repack/unmerged-%s" % fset)
            unmergedDataTier.loadData()
            dataTierMergeWorkflow = Workflow(name="TestWorkload",
                                             task="/TestWorkload/Repack/RepackMerge%s" % unmergedOutput)
            dataTierMergeWorkflow.load()
            mergeSubscription = Subscription(fileset=unmergedDataTier, workflow=dataTierMergeWorkflow)
            mergeSubscription.loadData()

            self.assertEqual(mergeSubscription["type"], "Merge",
                             "Error: Wrong subscription type.")
            self.assertEqual(mergeSubscription["split_algo"], "RepackMerge",
                             "Error: Wrong split algorithm. %s" % mergeSubscription["split_algo"])

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            fset = goldenOutputMod + tier
            unmergedFileset = Fileset(name="/TestWorkload/Repack/unmerged-%s" % fset)
            unmergedFileset.loadData()
            cleanupWorkflow = Workflow(name="TestWorkload",
                                       task="/TestWorkload/Repack/RepackCleanupUnmerged%s" % goldenOutputMod)
            cleanupWorkflow.load()
            cleanupSubscription = Subscription(fileset=unmergedFileset, workflow=cleanupWorkflow)
            cleanupSubscription.loadData()

            self.assertEqual(cleanupSubscription["type"], "Cleanup",
                             "Error: Wrong subscription type.")
            self.assertEqual(cleanupSubscription["split_algo"], "SiblingProcessingBased",
                             "Error: Wrong subscription type.")

        repackLogCollect = Fileset(name="/TestWorkload/Repack/unmerged-logArchive")
        repackLogCollect.loadData()
        repackLogCollectWorkflow = Workflow(name="TestWorkload",
                                            task="/TestWorkload/Repack/LogCollect")
        repackLogCollectWorkflow.load()
        logCollectSub = Subscription(fileset=repackLogCollect, workflow=repackLogCollectWorkflow)
        logCollectSub.loadData()

        self.assertEqual(logCollectSub["type"], "LogCollect",
                         "Error: Wrong subscription type.")
        self.assertEqual(logCollectSub["split_algo"], "MinFileBased",
                         "Error: Wrong split algorithm.")

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            repackMergeLogCollect = Fileset(
                name="/TestWorkload/Repack/RepackMerge%s/merged-logArchive" % goldenOutputMod)
            repackMergeLogCollect.loadData()
            repackMergeLogCollectWorkflow = Workflow(name="TestWorkload",
                                                     task="/TestWorkload/Repack/RepackMerge%s/Repack%sMergeLogCollect" % (
                                                         goldenOutputMod, goldenOutputMod))
            repackMergeLogCollectWorkflow.load()
            logCollectSubscription = Subscription(fileset=repackMergeLogCollect, workflow=repackMergeLogCollectWorkflow)
            logCollectSubscription.loadData()

            self.assertEqual(logCollectSub["type"], "LogCollect",
                             "Error: Wrong subscription type.")
            self.assertEqual(logCollectSub["split_algo"], "MinFileBased",
                             "Error: Wrong split algorithm.")

        return

    def testMemCoresSettings(self):
        """
        _testMemCoresSettings_
        
        Make sure the multicore and memory setings are properly propagated to
        all tasks and steps.
        """
        testArguments = RepackWorkloadFactory.getTestArguments()
        testArguments.update(deepcopy(REQUEST))

        factory = RepackWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", testArguments)

        # test default values
        taskPaths = ['/TestWorkload/Repack']
        for task in taskPaths:
            taskObj = testWorkload.getTaskByPath(task)
            for step in ('cmsRun1', 'stageOut1', 'logArch1'):
                stepHelper = taskObj.getStepHelper(step)
                self.assertEqual(stepHelper.getNumberOfCores(), 1)
                self.assertEqual(stepHelper.getNumberOfStreams(), 0)
            # then test Memory requirements
            perfParams = taskObj.jobSplittingParameters()['performance']
            self.assertEqual(perfParams['memoryRequirement'], 2300.0)

        # now test case where args are provided
        testArguments["Multicore"] = 6
        testArguments["Memory"] = 4600.0
        testArguments["EventStreams"] = 3
        testArguments["Outputs"] = deepcopy(REQUEST['Outputs'])

        factory = RepackWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", testArguments)

        for task in taskPaths:
            taskObj = testWorkload.getTaskByPath(task)
            for step in ('cmsRun1', 'stageOut1', 'logArch1'):
                stepHelper = taskObj.getStepHelper(step)
                if task == '/TestWorkload/Repack' and step == 'cmsRun1':
                    self.assertEqual(stepHelper.getNumberOfCores(), testArguments["Multicore"])
                    self.assertEqual(stepHelper.getNumberOfStreams(), testArguments["EventStreams"])
                elif step in ('stageOut1', 'logArch1'):
                    self.assertEqual(stepHelper.getNumberOfCores(), 1)
                    self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                else:
                    self.assertEqual(stepHelper.getNumberOfCores(), 1, "%s should be single-core" % task)
                    self.assertEqual(stepHelper.getNumberOfStreams(), 0)
            # then test Memory requirements
            perfParams = taskObj.jobSplittingParameters()['performance']
            self.assertEqual(perfParams['memoryRequirement'], testArguments["Memory"])

        return

    def testFilesets(self):
        """
        Test workflow tasks, filesets and subscriptions creation
        """
        # expected tasks, filesets, subscriptions, etc
        expOutTasks = ['/TestWorkload/Repack',
                       '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW',
                       '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW']
        expWfTasks = ['/TestWorkload/Repack',
                      '/TestWorkload/Repack/LogCollect',
                      '/TestWorkload/Repack/RepackCleanupUnmergedwrite_PrimaryDataset1_RAW',
                      '/TestWorkload/Repack/RepackCleanupUnmergedwrite_PrimaryDataset2_RAW',
                      '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW',
                      '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW/Repackwrite_PrimaryDataset1_RAWMergeLogCollect',
                      '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW',
                      '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW/Repackwrite_PrimaryDataset2_RAWMergeLogCollect']
        expFsets = ['TestWorkload-Repack-StreamerFiles',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW/merged-logArchive',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW/merged-MergedRAW',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW/merged-MergedErrorRAW',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW/merged-logArchive',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW/merged-MergedRAW',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW/merged-MergedErrorRAW',
                    '/TestWorkload/Repack/unmerged-write_PrimaryDataset1_RAWRAW',
                    '/TestWorkload/Repack/unmerged-write_PrimaryDataset2_RAWRAW',
                    '/TestWorkload/Repack/unmerged-logArchive']
        subMaps = [(4,
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW/merged-logArchive',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW/Repackwrite_PrimaryDataset1_RAWMergeLogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (7,
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW/merged-logArchive',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW/Repackwrite_PrimaryDataset2_RAWMergeLogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (8,
                    '/TestWorkload/Repack/unmerged-logArchive',
                    '/TestWorkload/Repack/LogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (2,
                    '/TestWorkload/Repack/unmerged-write_PrimaryDataset1_RAWRAW',
                    '/TestWorkload/Repack/RepackCleanupUnmergedwrite_PrimaryDataset1_RAW',
                    'SiblingProcessingBased',
                    'Cleanup'),
                   (3,
                    '/TestWorkload/Repack/unmerged-write_PrimaryDataset1_RAWRAW',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset1_RAW',
                    'RepackMerge',
                    'Merge'),
                   (5,
                    '/TestWorkload/Repack/unmerged-write_PrimaryDataset2_RAWRAW',
                    '/TestWorkload/Repack/RepackCleanupUnmergedwrite_PrimaryDataset2_RAW',
                    'SiblingProcessingBased',
                    'Cleanup'),
                   (6,
                    '/TestWorkload/Repack/unmerged-write_PrimaryDataset2_RAWRAW',
                    '/TestWorkload/Repack/RepackMergewrite_PrimaryDataset2_RAW',
                    'RepackMerge',
                    'Merge'),
                   (1,
                    'TestWorkload-Repack-StreamerFiles',
                    '/TestWorkload/Repack',
                    'Repack',
                    'Repack')]

        testArguments = RepackWorkloadFactory.getTestArguments()
        testArguments.update(deepcopy(REQUEST))

        factory = RepackWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", testArguments)

        testWMBSHelper = WMBSHelper(testWorkload, "Repack", blockName='StreamerFiles',
                                    cachepath=self.testInit.testDir)
        testWMBSHelper.createTopLevelFileset()
        testWMBSHelper._createSubscriptionsInWMBS(testWMBSHelper.topLevelTask, testWMBSHelper.topLevelFileset)

        self.assertItemsEqual(testWorkload.listOutputProducingTasks(), expOutTasks)

        workflows = self.listTasksByWorkflow.execute(workflow="TestWorkload")
        self.assertItemsEqual([item['task'] for item in workflows], expWfTasks)

        # returns a tuple of id, name, open and last_update
        filesets = self.listFilesets.execute()
        self.assertItemsEqual([item[1] for item in filesets], expFsets)

        subscriptions = self.listSubsMapping.execute(workflow="TestWorkload", returnTuple=True)
        self.assertItemsEqual(subscriptions, subMaps)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
_SetRunningJobSlotsForSite_

Set the number of running job slots available at a given site.

Created on Fri Jun 22 14:07:23 2012

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class SetRunningJobSlotsForSite(DBFormatter):
    sql = "UPDATE wmbs_location SET running_slots = :slots WHERE site_name = :site"

    def execute(self, siteName, runningSlots, conn = None, transaction = False):
        self.dbi.processData(self.sql, {"site": siteName, "slots": runningSlots},
                             conn = conn, transaction = transaction)
        return
